"""
MIT License

Copyright (c) 2025 YamboPy team 
Copyright (c) 2023 Stephen M. Goodlett, Nathaniel L. Kitzmiller

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
# Adapted and modified from MolSym package
# MolSym: A Python package for handling symmetry in molecular quantum chemistry
## Cite the following paper :https://doi.org/10.1063/5.0216738

import numpy as np
import re
from numpy.linalg import matrix_power
from dataclasses import dataclass
from scipy.spatial import KDTree
from yambopy.tools.citations import citation

global_tol = 1e-6


@citation("S. M. Goodlett et al. J. Chem. Phys. 161, 024107 (2024)")
def get_pg_info(symm_mats):
    """
    Given list of symmetries that form a crystallographic point 
    group, return point-group label, classes, character table, irrep_labels
    """
    order = len(symm_mats)
    pg_label = get_point_grp(symm_mats)
    pg_symels = pg_to_symels(pg_label)
    pg_sym_mats = []
    for i in pg_symels:
        pg_sym_mats.append(i.rrep)
    pg_sym_mats = np.array(pg_sym_mats)
    assert len(pg_sym_mats) == order
    transfom_mat = transform_matrix(symm_mats, pg_sym_mats)
    symm_mats_transformed = transfom_mat[
        None, :, :] @ symm_mats @ transfom_mat.T[None, :, :]
    assert np.imag(symm_mats_transformed).max() < 1e-4
    symm_mats_transformed = symm_mats_transformed.real
    sym_tree = KDTree(pg_sym_mats.reshape(order, -1))
    distance, idx = sym_tree.query(symm_mats_transformed.reshape(order, -1),
                                   k=1)
    assert np.max(distance) < 1e-4, np.max(distance)
    assert len(np.unique(idx)) == order
    ctab = pg_to_chartab(pg_label)
    classes = ctab.classes
    irreps = ctab.irreps
    char_tab = ctab.characters
    pg_class_map = np.array(generate_symel_to_class_map(pg_symels, ctab))
    class_map = pg_class_map[idx]
    class_dict = dict()
    for i, n in enumerate(class_map):
        class_dict.setdefault(n, []).append(i)
    class_dict = {n: rep for n, rep in class_dict.items() if len(rep) > 0}
    return pg_label, classes, class_dict, char_tab, irreps


def decompose_rep2irrep(red_rep, char_table, pg_order, class_order,
                        irrep_labels, tol=1e-3):
    irrep_coeff = np.einsum(
        'j,j,rj->r', class_order, red_rep, char_table, optimize=True) / pg_order
    assert np.abs(irrep_coeff.imag).max() < tol, print(
        np.abs(irrep_coeff.imag).max())
    irrep_coeff = irrep_coeff.real
    assert np.abs(irrep_coeff-np.rint(irrep_coeff)).max() < tol,\
                np.abs(irrep_coeff-np.rint(irrep_coeff)).max()
    irrep_coeff = np.rint(irrep_coeff).astype(int)
    rep_string = ''
    for i in range(len(irrep_labels)):
        if irrep_coeff[i] == 0:
            continue
        num_str = ''
        if irrep_coeff[i] > 1:
            num_str += str(irrep_coeff[i])
        rep_string = rep_string + num_str + irrep_labels[i] + ' + '
    return rep_string.strip()[:-1]


## Helpers
################################################################################
def normalize(a):
    """
    Normalize vector a to unit length
    :param a: Vector of arbitrary magnitude
    :type a: NumPy array of shape (n,)
    :return: Normalized vector or None if the magnitude of a is less than the global tolerance
    :rtype: NumPy array of shape (n,) or None
    """
    n = np.linalg.norm(a)
    if n <= global_tol:
        return a
    return a / np.linalg.norm(a)


def rotation_matrix(axis, theta):
    """
    Rotation matrix about an axis by theta in radians.

    :param axis: Cartesian vector defining rotation axis
    :param theta: Angle of rotation in radians
    :type axis: NumPy array of shape (3,)
    :type theta: float
    :return: Matrix defining rotation on column vector
    :rtype: NumPy array of shape (3,3)
    """
    cos_t = np.cos(theta)
    sin_t = np.sin(theta)
    # NOT NORMALIZING AXIS!!!
    M = cos_t * np.eye(3)
    M += sin_t * np.cross(np.eye(3), axis)
    M += (1 - cos_t) * np.outer(axis, axis)
    return M


def reflection_matrix(axis):
    """
    Reflection matrix about a plane defined by its normal vector.

    :param axis: Cartesian vector defining the plane normal vector
    :type axis: NumPy array of shape (3,)
    :return: Matrix defining reflection on column vector
    :rtype: NumPy array of shape (3,3)
    """
    M = np.zeros((3, 3))
    for i in range(3):
        for j in range(i, 3):
            if i == j:
                M[i, i] = 1 - 2 * (axis[i]**2)
            else:
                M[i, j] = -2 * axis[i] * axis[j]
                M[j, i] = M[i, j]
    return M


def inversion_matrix():
    """
    Cartesian inversion matrix.

    :return: Matrix defining inversion
    :rtype: NumPy array of shape(3,3)
    """
    return -1 * np.eye(3)


def Cn(axis, n):
    """
    Wrapper around rotation_matrix for producing a C_n rotation about axis.

    :param axis: Cartesian vector defining rotation axis
    :param n: Defines rotation angle by theta = 2 pi / n
    :type axis: NumPy array of shape (3,)
    :type n: int
    :return: Matrix defining proper rotation on column vector
    :rtype: NumPy array of shape (3,3)
    """
    theta = 2 * np.pi / n
    return rotation_matrix(axis, theta)


def Sn(axis, n):
    """
    Improper rotation S_n about an axis.
    
    :param axis: Cartesian vector defining rotation axis
    :param n: Defines rotation angle by theta = 2 pi / n
    :type axis: NumPy array of shape (3,)
    :type n: int
    :return: Matrix defining improper rotation on column vector
    :rtype: NumPy array of shape (3,3)
    """
    return np.dot(reflection_matrix(axis), Cn(axis, n))

def get_symmetry_axes(mats, tol=1e-3):
    """
    Extracts unique rotation axes and their orders.
    Robustly handles float32 noise and det scaling.
    """
    unique_axes = [] # List of [order, axis]
    
    for mat in mats:
        # 1. Use sign of determinant to handle Improper Rotations without scaling magnitude
        det = np.linalg.det(mat)
        sign = 1.0 if det > 0 else -1.0
        R_part = mat * sign
        
        # 2. Eigen decomposition
        w, v = np.linalg.eig(R_part)
        
        # 3. Find eigenvalue ~ 1.0
        idx = np.argmin(np.abs(w - 1.0))
        if np.abs(w[idx] - 1.0) > tol: continue
            
        # 4. Extract Real Axis Robustly
        # Handle case where eig returns complex vector for real eigenvalue
        vec = v[:, idx]
        if np.all(np.abs(vec.imag) < tol):
            axis = np.real(vec)
        else:
            # Try to rotate phase to make it real (unlikely for valid R_part, but safe)
            # or just take real part if imag is small noise
            axis = np.real(vec)
            
        norm = np.linalg.norm(axis)
        if norm < tol: continue
        axis = axis / norm
        
        # 5. Enforce unique hemisphere
        for k in range(3):
            if abs(axis[k]) > tol:
                if axis[k] < 0: axis = -axis
                break

        # 6. Calculate Order
        tr = np.trace(R_part)
        cos_t = np.clip((tr - 1.0) / 2.0, -1.0, 1.0)
        theta = np.arccos(cos_t)
        
        if abs(theta) < tol: continue # Identity
            
        n = int(np.round(2 * np.pi / theta))
        if n < 2: continue

        # 7. Check Uniqueness & Upgrade Order
        found = False
        for i in range(len(unique_axes)):
            old_n, old_ax = unique_axes[i]
            if 1.0 - abs(np.dot(axis, old_ax)) < tol:
                found = True
                if n > old_n:
                    unique_axes[i] = [n, axis]
                break
        
        if not found:
            unique_axes.append([n, axis])
            
    unique_axes.sort(key=lambda x: x[0], reverse=True)
    return [tuple(x) for x in unique_axes]

def align_basis(u1, v1, u2, v2):
    """Constructs rotation matrix R aligning (u1, v1) to (u2, v2)."""
    # Frame 1
    x1 = u1 / np.linalg.norm(u1)
    y1 = v1 - np.dot(v1, x1) * x1
    if np.linalg.norm(y1) < 1e-4: return None
    y1 = y1 / np.linalg.norm(y1)
    z1 = np.cross(x1, y1)
    F1 = np.column_stack((x1, y1, z1))
    
    # Frame 2
    x2 = u2 / np.linalg.norm(u2)
    y2 = v2 - np.dot(v2, x2) * x2
    if np.linalg.norm(y2) < 1e-4: return None
    y2 = y2 / np.linalg.norm(y2)
    z2 = np.cross(x2, y2)
    F2 = np.column_stack((x2, y2, z2))
    
    return F2 @ F1.T

def check_transform(R, mats1, mats2, tol=1e-3):
    """Verifies if R @ mats1 @ R.T is a subset of mats2."""
    for m1 in mats1:
        m1_trans = R @ m1 @ R.T
        found = False
        # Optimization: Filter by trace first (rotation angle invariant)
        tr1 = np.trace(m1_trans)
        for m2 in mats2:
            if abs(tr1 - np.trace(m2)) > tol: continue
            if np.allclose(m1_trans, m2, atol=tol):
                found = True
                break
        if not found: return False
    return True

def transform_matrix(sym_mats_old, sym_mats_new):
    """
    Finds transformation R mapping sym_mats_old to sym_mats_new.
    Uses Geometric Hashing and Relaxed Order Checks.
    """
    mats1 = np.array(sym_mats_old)[:, :3, :3]
    mats2 = np.array(sym_mats_new)[:, :3, :3]

    if len(mats1) <= 1: return np.eye(3)

    axes1 = get_symmetry_axes(mats1)
    axes2 = get_symmetry_axes(mats2)

    if not axes1 or not axes2: return np.eye(3)

    # Strategy 1: Two-Vector Alignment
    n1_a, u1 = axes1[0]

    v1 = None
    for n, ax in axes1[1:]:
        if abs(np.dot(u1, ax)) < 1.0 - 1e-3:
            v1 = ax
            break

    if v1 is not None:
        target_angle = np.arccos(np.clip(np.dot(u1, v1), -1.0, 1.0))

        # Search all compatible pairs in target group
        for n2_a, u2 in axes2:
            # RELAXED CHECK: Iterate all u2 to handle order mismatches (e.g. C3 vs C6)

            for n2_b, v2 in axes2:
                # Check angle match
                ang = np.arccos(np.clip(np.dot(u2, v2), -1.0, 1.0))
                if abs(ang - target_angle) > 1e-3: continue

                R = align_basis(u1, v1, u2, v2)
                if R is not None and check_transform(R, mats1, mats2):
                    return R

    # Strategy 2: Single Axis Alignment
    for n2_a, u2 in axes2:
        # RELAXED CHECK: Try aligning primary axes even if orders mismatch

        dot = np.dot(u1, u2)
        if dot > 1.0 - 1e-4:
            R = np.eye(3)
        elif dot < -1.0 + 1e-4:
            perp = np.array([1.0, 0.0, 0.0])
            if abs(np.dot(perp, u1)) > 0.9: perp = np.array([0.0, 1.0, 0.0])
            rot_ax = np.cross(u1, perp)
            rot_ax /= np.linalg.norm(rot_ax)
            R = rotation_matrix(rot_ax, np.pi)
        else:
            rot_ax = np.cross(u1, u2)
            rot_ax /= np.linalg.norm(rot_ax)
            angle = np.arccos(np.clip(dot, -1.0, 1.0))
            R = rotation_matrix(rot_ax, angle)

        if check_transform(R, mats1, mats2):
            return R

    return np.eye(3)


def reduce(n, i):
    """
    Divide n and i by their greatest common divisor g.

    :type n: int
    :type i: int
    :return: Tuple of n/g and i/g
    :rtype: (int, int)
    """
    g = gcd(n, i)
    return n // g, i // g
    # floor divide to get an int, there should never be a remainder since we are dividing by the gcd


def gcd(A, B):
    """
    A quick implementation of the Euclid algorithm for finding the greatest common divisor between A and B.
    
    :type A: int
    :type B: int
    :return: Greatest common divisor between A and B
    :rtype: int
    """
    a = max(A, B)
    b = min(A, B)
    if a == 0:
        return b
    elif b == 0:
        return a
    else:
        r = a % b
        return gcd(b, r)

class PointGroup():
    """
    Class for defining point group.
    """

    def __init__(self, s, family, n, subfamily):
        self.str = s
        self.family = family
        self.n = n
        if self.n == 0:
            self.is_linear = True
        else:
            self.is_linear = False
        self.subfamily = subfamily
        self.dumb_pg()

    @classmethod
    def from_string(cls, s):
        regex = r"([A-Z]+)(\d+)?([a-z]+)?"
        m = re.match(regex, s)
        family, n, subfamily = m.groups()
        if n is not None:
            n = int(n)
        if subfamily is not None:
            subfamily = str(subfamily)
        family = str(family)
        return cls(s, family, n, subfamily)

    def __str__(self):
        nstr = self.n
        sfstr = self.subfamily
        if self.n is None:
            nstr = ""
        elif self.n == 0:
            nstr = "inf"
        else:
            nstr = str(self.n)
        if self.subfamily is None:
            sfstr = ""
        return self.family + nstr + sfstr

    def __repr__(self) -> str:
        return self.__str__()

    def dumb_pg(self):
        # Check if a dumb point group has been made (e.g. D1h, D0v, C2i)
        argstr = (f"You have generated a dumb point group:" +
                  " {self.str}. Family {self.family}, n {self.n}," +
                  " subfamily {self.subfamily}. We aren't sure " +
                  "how you managed to do this but we aren't paid " +
                  "enough to proceed with any calculations. " +
                  "If you have any questions, " +
                  "feel free to email the CFOUR listserv.")
        if self.n is None:
            if self.family == "C":
                allowed = ["s", "i"]
                if self.subfamily in allowed:
                    return 0
            elif self.family == "T":
                allowed = [None, "h", "d"]
                if self.subfamily in allowed:
                    return 0
            elif self.family == "O" or self.family == "I":
                allowed = [None, "h"]
                if self.subfamily in allowed:
                    return 0
        elif self.n == 0:
            if self.family == "D" and self.subfamily == "h":
                return 0
            elif self.family == "C" and self.subfamily == "v":
                return 0
        elif self.n == 1:
            if self.family == "C" and self.subfamily is None:
                return 0
        elif self.n >= 2:
            if self.family == "C":
                allowed = [None, "v", "h"]
                if self.subfamily in allowed:
                    return 0
            elif self.family == "D":
                allowed = [None, "d", "h"]
                if self.subfamily in allowed:
                    return 0
            elif self.family == "S":
                if self.subfamily is None and self.n % 2 == 0:
                    return 0
        raise Exception(argstr)


class CharacterTable():

    def __init__(self, pg, irreps, classes, class_orders, chars,
                 irrep_dims) -> None:
        self.name = pg
        self.irreps = irreps
        self.classes = classes
        self.class_orders = class_orders
        self.characters = chars
        self.irrep_dims = irrep_dims

    def __repr__(self) -> str:
        return f"Character Table for {self.name}\nIrreps: \
                {self.irreps}\nClasses: {self.classes}\nCharacters:\n{self.characters}\n"

    def __eq__(self, other):
        if len(self.irreps) == len(other.irreps) and \
                len(self.classes) == len(other.classes) and \
                np.shape(self.characters)==np.shape(other.characters):
            return (self.irreps == other.irreps).all() and \
                    (self.classes == other.classes).all() and \
                    np.isclose(self.characters,other.characters,atol=1e-10).all()
        else:
            return False


def pg_to_chartab(PG):
    pg = PointGroup.from_string(PG)
    irreps = []
    if pg.family == "C":
        if pg.subfamily == "s":
            irreps = ["A'", "A''"]
            classes = ["E", "sigma_h"]
            chars = np.array([[1.0, 1.0], [1.0, -1.0]])
        elif pg.subfamily == "i":
            irreps = ["Ag", "Au"]
            classes = ["E", "i"]
            chars = np.array([[1.0, 1.0], [1.0, -1.0]])
        elif pg.subfamily == "v":
            irreps, classes, chars = Cnv_irr(pg.n)
        elif pg.subfamily == "h":
            irreps, classes, chars = Cnh_irr(pg.n)
        else:
            #irreps, classes, chars = Cn_irrmat(pg.n)
            irreps, classes, chars = Cn_irr_complex(pg.n)
    elif pg.family == "D":
        if pg.subfamily == "d":
            irreps, classes, chars = Dnd_irr(pg.n)
        elif pg.subfamily == "h":
            irreps, classes, chars = Dnh_irr(pg.n)
        else:
            irreps, classes, chars = Dn_irr(pg.n)
    elif pg.family == "S":
        irreps, classes, chars = Sn_irr_complex(pg.n)
    else:
        cp3 = np.cos(np.pi / 3)
        pr5 = 0.5 * (1.0 + np.sqrt(5.0))
        mr5 = 0.5 * (1.0 - np.sqrt(5.0))
        if pg.family == "T":
            if pg.subfamily == "h":
                irreps, classes, chars = (
                    ["Ag", "Au", "Eg", "Eu", "Tg", "Tu"], [
                        "E", "4C_3", "4C_3^2", "3C_2", "i", "S_6", "S_6^5",
                        "3sigma_h"
                    ],
                    np.array([[1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                              [1.0, 1.0, 1.0, 1.0, -1.0, -1.0, -1.0, -1.0],
                              [2.0, cp3, cp3, 2.0, 2.0, cp3, cp3, 1.0],
                              [2.0, cp3, cp3, 2.0, -2.0, -cp3, -cp3, -1.0],
                              [3.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, -1.0],
                              [3.0, 0.0, 0.0, -1.0, -1.0, 0.0, 0.0, 1.0]]))
            elif pg.subfamily == "d":
                irreps, classes, chars = (["A1", "A2", "E", "T1", "T2"], [
                    "E", "8C_3", "3C_2", "6S_4", "6sigma_d"
                ],
                                          np.array([[1.0, 1.0, 1.0, 1.0, 1.0],
                                                    [1.0, 1.0, 1.0, -1.0, -1.0],
                                                    [2.0, -1.0, 2.0, 0.0, 0.0],
                                                    [3.0, 0.0, -1.0, 1.0, -1.0],
                                                    [3.0, 0.0, -1.0, -1.0,
                                                     1.0]]))
            else:
                irreps, classes, chars = (["A", "E", "T"],
                                          ["E", "4C_3", "4C_3^2", "3C_2"],
                                          np.array([[1.0, 1.0, 1.0, 1.0],
                                                    [2.0, cp3, cp3, 2.0],
                                                    [3.0, 0.0, 0.0, -1.0]]))
        elif pg.family == "O":
            if pg.subfamily == "h":
                irreps, classes, chars = (
                    [
                        "A1g", "A2g", "Eg", "T1g", "T2g", "A1u", "A2u", "Eu",
                        "T1u", "T2u"
                    ], [
                        "E", "8C_3", "6C_2", "6C_4", "3C_2", "i", "6S_4",
                        "8S_6", "3sigma_h", "6sigma_d"
                    ],
                    np.array([
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [1.0, 1.0, -1.0, -1.0, 1.0, 1.0, -1.0, 1.0, 1.0, -1.0],
                        [2.0, -1.0, 0.0, 0.0, 2.0, 2.0, 0.0, -1.0, 2.0, 0.0],
                        [3.0, 0.0, -1.0, 1.0, -1.0, 3.0, 1.0, 0.0, -1.0, -1.0],
                        [3.0, 0.0, 1.0, -1.0, -1.0, 3.0, -1.0, 0.0, -1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, -1.0, -1.0, -1.0, -1.0, -1.0],
                        [1.0, 1.0, -1.0, -1.0, 1.0, -1.0, 1.0, -1.0, -1.0, 1.0],
                        [2.0, -1.0, 0.0, 0.0, 2.0, -2.0, 0.0, 1.0, -2.0, 0.0],
                        [3.0, 0.0, -1.0, 1.0, -1.0, -3.0, -1.0, 0.0, 1.0, 1.0],
                        [3.0, 0.0, 1.0, -1.0, -1.0, -3.0, 1.0, 0.0, 1.0, -1.0]
                    ]))
            else:
                irreps, classes, chars = (["A1", "A2", "E", "T1", "T2"],
                                          ["E", "6C_4", "3C_2", "8C_3", "6C_2"],
                                          np.array([[1.0, 1.0, 1.0, 1.0, 1.0],
                                                    [1.0, -1.0, 1.0, 1.0, -1.0],
                                                    [2.0, 0.0, 2.0, -1.0, 0.0],
                                                    [3.0, 1.0, -1.0, 0.0, -1.0],
                                                    [3.0, -1.0, -1.0, 0.0,
                                                     1.0]]))
        elif pg.family == "I":
            if pg.subfamily == "h":
                irreps, classes, chars = (
                    [
                        "Ag", "T1g", "T2g", "Gg", "Hg", "Au", "T1u", "T2u",
                        "Gu", "Hu"
                    ], [
                        "E", "12C_5", "12C_5^2", "20C_3", "15C_2", "i",
                        "12S_10", "12S_10^3", "20S_6", "15sigma"
                    ],
                    np.array([
                        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                        [3.0, pr5, mr5, 0.0, -1.0, 3.0, mr5, pr5, 0.0, -1.0],
                        [3.0, mr5, pr5, 0.0, -1.0, 3.0, pr5, mr5, 0.0, -1.0],
                        [4.0, -1.0, -1.0, 1.0, 0.0, 4.0, -1.0, -1.0, 1.0, 0.0],
                        [5.0, 0.0, 0.0, -1.0, 1.0, 5.0, 0.0, 0.0, -1.0, 1.0],
                        [1.0, 1.0, 1.0, 1.0, 1.0, -1.0, -1.0, -1.0, -1.0, -1.0],
                        [3.0, pr5, mr5, 0.0, -1.0, -3.0, -mr5, -pr5, 0.0, 1.0],
                        [3.0, mr5, pr5, 0.0, -1.0, -3.0, -pr5, -mr5, 0.0, 1.0],
                        [4.0, -1.0, -1.0, 1.0, 0.0, -4.0, 1.0, 1.0, -1.0, 0.0],
                        [5.0, 0.0, 0.0, -1.0, 1.0, -5.0, 0.0, 0.0, 1.0, -1.0]
                    ]))
            else:
                irreps, classes, chars = (["A", "T1", "T2", "G", "H"], [
                    "E", "12C_5", "12C_5^2", "20C_3", "15C_2"
                ],
                                          np.array([[1.0, 1.0, 1.0, 1.0, 1.0],
                                                    [3.0, pr5, mr5, 0.0, -1.0],
                                                    [3.0, mr5, pr5, 0.0, -1.0],
                                                    [4.0, -1.0, -1.0, 1.0, 0.0],
                                                    [5.0, 0.0, 0.0, -1.0,
                                                     1.0]]))
        else:
            raise Exception(f"An invalid point group has been " +
                            "given or unexpected parsing of the " +
                            "point group string has occured: {pg.str}")
    class_orders = grab_class_orders(classes)
    irr_dims = {}
    for (irr_idx, irrep) in enumerate(irreps):
        if pg.n == 1:
            irr_dims[irrep] = int(chars[0])
        else:
            irr_dims[irrep] = int(np.real(chars[irr_idx, 0]))
    return CharacterTable(PG, irreps, classes, class_orders, chars, irr_dims)


def grab_class_orders(classes):
    ncls = len(classes)
    class_orders = np.zeros(ncls)
    for i in range(ncls):  # = 1:ncls
        class_orders[i] = grab_order(classes[i])
    return class_orders


def grab_order(class_str):
    regex = r"^(\d+)"
    m = re.match(regex, class_str)
    if m is not None:
        return int(m.groups()[0])
    else:
        return 1


def c_class_string(classes, pre, r, s):
    "Pushes class string to 'classes' for rotations, but ignores superscript if s is one"
    if s == 1:
        classes.append(pre + f"_{r}")
    else:
        classes.append(pre + f"_{r}^{s}")


def Cn_irrmat(n):
    names = ["A"]
    classes = ["E"]
    for c in range(1, n):
        r, s = reduce(n, c)
        c_class_string(classes, "C", r, s)
    chars = np.ones(n)
    if n % 2 == 0:
        names.append("B")
        bi = np.ones(n)
        #for i=1:n:
        for i in range(n):
            if (i + 1) % 2 == 0:
                bi[i] *= -1
        chars = np.vstack((chars, bi))
    if 2 < n < 5:
        # No label associated with E if n < 5
        names.append("E")
        theta = 2 * np.pi / n
        v = np.zeros(n)
        for j in range(n):
            v[j] += 2 * np.cos(j * theta)
        chars = np.vstack((chars, v))
    elif n >= 5:
        theta = 2 * np.pi / n
        l = round(((n - len(names)) / 2))
        for i in range(l):  #= 1:l:
            names.append(f"E{i+1}")
            v = np.zeros(n)
            for j in range(n):
                v[j] += 2 * np.cos((i + 1) * j * theta)
            chars = np.vstack((chars, v))
    return names, classes, chars


def Cn_irr_complex(n):
    names = ["A"]
    classes = ["E"]
    for c in range(1, n):
        r, s = reduce(n, c)
        c_class_string(classes, "C", r, s)
    chars = np.ones(n)
    if n % 2 == 0:
        names.append("B")
        bi = np.ones(n)
        #for i=1:n:
        for i in range(n):
            if (i + 1) % 2 == 0:
                bi[i] *= -1
        chars = np.vstack((chars, bi))
    if 2 < n < 5:
        # No label associated with E if n < 5
        names.append("E_1")
        names.append("E_2")
        theta = np.exp(2 * np.pi * 1j / n)
        v1 = np.zeros(n, dtype=np.complex128)
        v2 = np.zeros(n, dtype=np.complex128)
        for a in range(n):
            v1[a] += theta**a
            v2[a] += np.conj(theta**a)
        chars = np.vstack((chars, v1))
        chars = np.vstack((chars, v2))
    elif n >= 5:
        theta = 2 * np.pi / n
        l = round(((n - len(names)) / 2))
        for a in range(l):  #= 1:l:
            names.append(f"E{a+1}_1")
            names.append(f"E{a+1}_2")
            theta = np.exp(2 * np.pi * 1j * (a + 1) / n)
            v1 = np.zeros(n, dtype=np.complex128)
            v2 = np.zeros(n, dtype=np.complex128)
            for b in range(n):
                v1[b] += theta**b
                v2[b] += np.conj(theta**b)
            chars = np.vstack((chars, v1))
            chars = np.vstack((chars, v2))
    return names, classes, chars


def Cnv_irr(n):
    names, classes, chars = Cn_irrmat(n)
    classes = ["E"]
    if n % 2 == 0:
        for c in range(1, n >> 1):  # = 1:(n>>1)-1:
            r, s = reduce(n, c)
            c_class_string(classes, "2C", r, s)
        classes.append("C_2")
        if n >> 1 == 1:
            classes.append("sigma_v(xz)")
            classes.append("sigma_d(yz)")
        else:
            classes.append(f"{n>>1}sigma_v")
            classes.append(f"{n>>1}sigma_d")
    else:
        for c in range(1, (n >> 1) + 1):  # = 1:n>>1:
            r, s = reduce(n, c)
            c_class_string(classes, "2C", r, s)
        classes.append(f"{n}sigma_v")
    names[0] = "A1"
    names.insert(1, "A2")
    chars = np.vstack((chars[0, :], chars[0, :], chars[1:, :]))
    for i in range(1, n):  # = 2:n:
        if i >= n - i:
            break
        #deleteat!(classes, n-i+2)
        chars = chars[:, [j for j in range(np.shape(chars)[1]) if j != n - i
                         ]]  #chars[:,1:-1.!=n-i+2]
    if n % 2 == 0:
        nirr = round((n / 2) + 3)
        names[2] = "B1"
        names.insert(3, "B2")
        chars = np.vstack((chars[0:3, :], chars[2, :], chars[3:, :]))
        sigma_v = np.zeros(nirr)
        sigma_d = np.zeros(nirr)
        sigma_v[0:4] = np.array([1, -1, 1, -1])
        sigma_d[0:4] = np.array([1, -1, -1, 1])
        chars = np.hstack((chars, sigma_v[:, None], sigma_d[:, None]))
    else:
        nirr = round((n - 1) / 2 + 2)
        sigma_v = np.zeros(nirr)
        sigma_v[0:2] = np.array([1, -1])
        chars = np.hstack((chars, sigma_v[:, None]))
    return names, classes, chars


def Cnh_irr(n):
    #names, classes, cnchars = Cn_irrmat(n)
    names, classes, cnchars = Cn_irr_complex(n)
    if n % 2 == 0:
        classes.append("i")
        for i in range(1, n):  # = 1:n-1:
            if i == n >> 1:
                classes.append("sigma_h")
            else:
                r, s = reduce(n, (i + (n >> 1)) % n)
                if s % 2 == 0:
                    s += r
                c_class_string(classes, "S", r, s)
    else:
        classes.append("sigma_h")
        for i in range(1, n):  # = 1:n-1:
            r, s = reduce(n, i)
            if i % 2 == 0:
                c_class_string(classes, "S", r, s + n)
            else:
                c_class_string(classes, "S", r, s)
    if n % 2 == 0:
        newnames = []
        for i in range(len(names)):  # = 1:length(names):
            newnames.append(names[i] + "u")
            names[i] = names[i] + "g"
        names += newnames
        cncharsi = -1 * cnchars
        top = np.hstack((cnchars, cnchars))
        bot = np.hstack((cnchars, cncharsi))
        chars = np.vstack((top, bot))
    else:
        newnames = []
        for i in range(len(names)):  # = 1:length(names):
            newnames.append(names[i] + "''")
            names[i] = names[i] + "'"
        names += newnames
        cncharsi = -1 * cnchars
        top = np.hstack((cnchars, cnchars))
        bot = np.hstack((cnchars, cncharsi))
        chars = np.vstack((top, bot))
    return names, classes, chars


def Sn_irr(n):
    if n % 4 == 0:
        names, classes, chars = Cn_irrmat(n)
        for i in range(n):  # = 1:n:
            if (i + 1) % 2 == 0:
                classes[i] = "S" + classes[i][1:]
    elif n % 2 == 0:
        ni = round(n / 2)
        names, classes, cnchars = Cn_irrmat(ni)
        classes = ["E"]
        for i in range(1, n >> 1):  # = 1:n>>1-1:
            r, s = reduce(n >> 1, i)
            c_class_string(classes, "C", r, s)
        classes.append("i")
        for i in range(1, n >> 1):  # = 1:n>>1-1:
            r, s = reduce(n, ((i << 1) + (n >> 1)) % n)
            c_class_string(classes, "S", r, s)
        newnames = []
        for i in range(len(names)):  # = 1:length(names):
            newnames.append(names[i] + "u")
            names[i] = names[i] + "g"
        names += newnames
        cncharsi = -1 * cnchars
        top = np.hstack((cnchars, cnchars))
        bot = np.hstack((cnchars, cncharsi))
        chars = np.vstack((top, bot))
    else:
        raise Exception("Odd number n for S group")
    return names, classes, chars


def Sn_irr_complex(n):
    if n % 4 == 0:
        names, classes, chars = Cn_irr_complex(n)
        for i in range(n):  # = 1:n:
            if (i + 1) % 2 == 0:
                classes[i] = "S" + classes[i][1:]
    elif n % 2 == 0:
        ni = round(n / 2)
        names, classes, cnchars = Cn_irr_complex(ni)
        classes = ["E"]
        for i in range(1, n >> 1):  # = 1:n>>1-1:
            r, s = reduce(n >> 1, i)
            c_class_string(classes, "C", r, s)
        classes.append("i")
        for i in range(1, n >> 1):  # = 1:n>>1-1:
            r, s = reduce(n, ((i << 1) + (n >> 1)) % n)
            c_class_string(classes, "S", r, s)
        newnames = []
        for i in range(len(names)):  # = 1:length(names):
            newnames.append(names[i] + "u")
            names[i] = names[i] + "g"
        names += newnames
        cncharsi = -1 * cnchars
        top = np.hstack((cnchars, cnchars))
        bot = np.hstack((cnchars, cncharsi))
        chars = np.vstack((top, bot))
    else:
        raise Exception("Odd number n for S group")
    return names, classes, chars


def Dn_irr(n):
    if n == 2:
        names = ["A", "B1", "B2", "B3"]
        classes = ["E", "C_2(z)", "C_2(y)", "C_2(x)"]
        chars = np.array([[1, 1, 1, 1], [1, 1, -1, -1], [1, -1, 1, -1],
                          [1, -1, -1, 1]])
        return names, classes, chars
    names, garbage_classes, chars = Cn_irrmat(n)
    garbage_names, classes, garbage_chars = Cnv_irr(n)
    if n % 2 == 0:
        classes[-2] = classes[-2][0] + "C_2'"
        classes[-1] = classes[-1][0] + "C_2''"
    else:
        classes[-1] = classes[-1][0] + "C_2"
    names[0] = "A1"
    names.insert(1, "A2")
    chars = np.vstack((chars[0, :], chars[0, :], chars[1:, :]))
    for i in range(1, n):  # = 2:n:
        if i == n - i or i > n - i:
            break
        #deleteat!(classes, n-i+2)
        chars = chars[:, [j for j in range(np.shape(chars)[1]) if j != n - i
                         ]]  # chars[:,0:-1.!=n-i+2]
    if n % 2 == 0:
        nirr = round((n / 2) + 3)
        names[2] = "B1"
        names.insert(3, "B2")
        chars = np.vstack((chars[0:3, :], chars[2, :], chars[3:, :]))
        C2p = np.zeros(nirr)
        C2pp = np.zeros(nirr)
        C2p[0:4] = np.array([1, -1, 1, -1])
        C2pp[0:4] = np.array([1, -1, -1, 1])
        chars = np.hstack((chars, C2p[:, None], C2pp[:, None]))
    else:
        nirr = round((n - 1) / 2 + 2)
        C2p = np.zeros(nirr)
        C2p[0:2] = np.array([1, -1])
        chars = np.hstack((chars, C2p[:, None]))
    return names, classes, chars


def Dnh_irr(n):
    names, classes, dnchars = Dn_irr(n)
    if n % 2 == 0:
        classes.append("i")
        if n == 2:
            classes.append("sigma(xy)")
            classes.append("sigma(xz)")
            classes.append("sigma(yz)")
        else:
            for i in range(1, n >> 1):  # = 1:n>>1-1:
                a = i + (n >> 1)
                if a > (n >> 1):
                    a = n - a
                r, s = reduce(n, a)
                c_class_string(classes, "2S", r, s)
            classes.append("sigma_h")
            if n % 4 == 0:
                classes.append(f"{n>>1}sigma_v")
                classes.append(f"{n>>1}sigma_d")
            else:
                classes.append(f"{n>>1}sigma_d")
                classes.append(f"{n>>1}sigma_v")
        newnames = []
        for i in range(len(names)):  # = 1:length(names):
            newnames.append(names[i] + "u")
            names[i] = names[i] + "g"
        names += newnames
        dncharsi = -1 * dnchars
        top = np.hstack((dnchars, dnchars))
        bot = np.hstack((dnchars, dncharsi))
        chars = np.vstack((top, bot))
    else:
        classes.append("sigma_h")
        for i in range(1, (n >> 1) + 1):  # = 1:n>>1:
            if i % 2 == 0:
                r, s = reduce(n, n - i)
            else:
                r, s = reduce(n, i)
            c_class_string(classes, "2S", r, s)
        classes.append(f"{n}sigma_v")
        newnames = []
        for i in range(len(names)):  # = 1:length(names):
            newnames.append(names[i] + "''")
            names[i] = names[i] + "'"
        names += newnames
        dncharsi = -1 * dnchars
        top = np.hstack((dnchars, dnchars))
        bot = np.hstack((dnchars, dncharsi))
        chars = np.vstack((top, bot))
    return names, classes, chars


def Dnd_irr(n):
    if n % 2 == 0:
        n2 = 2 * n
        names, classes, chars = Sn_irr(n2)
        #classes = collect(1:2*n2)
        classes = classes[0:n + 1]
        for i in range(1, n):  # = 2:n:
            classes[i] = "2" + classes[i]
        classes.append(f"{n}C_2'")
        classes.append(f"{n}sigma_d")
        names[0] = "A1"
        names.insert(1, "A2")
        chars = np.vstack((chars[0, :], chars[0, :], chars[1:, :]))
        for i in range(1, n2):  # = 2:n2:
            if i >= n2 - i:
                break
            chars = chars[:,
                          [j for j in range(np.shape(chars)[1]) if j != n2 - i
                          ]]  # chars[:,0:-1.!=n2-i+2]
        nirr = n + 3
        names[2] = "B1"
        names.insert(3, "B2")
        chars = np.vstack((chars[0:3, :], chars[2, :], chars[3:, :]))
        C2p = np.zeros(nirr)
        sigma_d = np.zeros(nirr)
        C2p[0:4] = np.array([1, -1, 1, -1])
        sigma_d[0:4] = np.array([1, -1, -1, 1])
        chars = np.hstack((chars, C2p[:, None], sigma_d[:, None]))
    else:
        names, classes, dnchars = Dn_irr(n)
        classes.append("i")
        for i in range(1, (n >> 1) + 1):  # = 1:n>>1:
            r, s = reduce(2 * n, 2 * i + n)
            if s > n:
                s = 2 * n - s
            c_class_string(classes, "2S", r, s)
        classes.append(f"{n}sigma_d")
        newnames = []
        for i in range(len(names)):  # = 1:length(names):
            newnames.append(names[i] + "u")
            names[i] = names[i] + "g"
        names += newnames
        dncharsi = -1 * dnchars
        top = np.hstack((dnchars, dnchars))
        bot = np.hstack((dnchars, dncharsi))
        chars = np.vstack((top, bot))
    return names, classes, chars


@dataclass
class Symel():
    """
    Deprecated data structure for symmetry elements.
    Still hanging around because of cubic/icosahedral groups and tests.
    """
    symbol: str
    vector: np.array  # Not defined for E or i, axis vector for Cn and Sn, plane normal vector for sigma
    rrep: np.array

    def __str__(self) -> str:
        with np.printoptions(precision=5,
                             suppress=True,
                             formatter={"all": lambda x: f"{x:8.5f}"}):
            return f"\nSymbol: {self.symbol:>10s}: [{self.rrep[0,:]},{self.rrep[1,:]},{self.rrep[2,:]}]"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        return self.symbol == other.symbol and np.isclose(
            self.rrep, other.rrep, atol=1e-10).all()


def pg_to_symels(PG):
    pg = PointGroup.from_string(PG)
    argerr = f"An invalid point group has been given or "+\
        "unexpected parsing of the point group string has occured: {pg.str}"
    symels = [Symel("E", None, np.asarray([[1, 0, 0], [0, 1, 0], [0, 0, 1]]))]
    z_axis = np.array([0, 0, 1])
    sigma_h = np.asarray([[1, 0, 0], [0, 1, 0], [0, 0, -1]])
    if pg.family == "C":
        if pg.subfamily == "h":
            symels.append(Symel("sigma_h", z_axis, sigma_h))
            if pg.n % 2 == 0:
                symels.append(Symel("i", None, inversion_matrix()))
            cns = generate_Cn(pg.n)
            sns = generate_Sn(pg.n)
            symels = symels + cns + sns
        elif pg.subfamily == "v":
            cns = generate_Cn(pg.n)
            if pg.n % 2 == 0:
                n = pg.n >> 1
                sigma_ds = generate_sigma_d(n)
            else:
                n = pg.n
                sigma_ds = []
            sigma_vs = generate_sigma_v(pg.n)
            symels = symels + cns + sigma_vs + sigma_ds
        elif pg.subfamily == "s":
            symels.append(Symel("sigma_h", z_axis, sigma_h))
        elif pg.subfamily == "i":
            symels.append(Symel("i", None, inversion_matrix()))
        elif pg.subfamily is None:
            cns = generate_Cn(pg.n)
            symels = symels + cns
        else:
            raise Exception(argerr)
    elif pg.family == "D":
        if pg.subfamily == "h":
            symels.append(Symel("sigma_h", z_axis, sigma_h))
            if pg.n % 2 == 0:
                symels.append(Symel("i", None, inversion_matrix()))
                n = pg.n >> 1
                sigma_ds = generate_sigma_d(n)
                c2ps = generate_C2p(pg.n)
                c2pps = generate_C2pp(pg.n)
                c2s = c2ps + c2pps
            else:
                n = pg.n
                sigma_ds = []
                c2s = generate_C2p(pg.n)
            cns = generate_Cn(pg.n)
            sns = generate_Sn(pg.n)
            sigma_vs = generate_sigma_v(pg.n)
            #c2s = generate_C2(pg.n)
            symels = symels + cns + c2s + sns + sigma_vs + sigma_ds
        elif pg.subfamily == "d":
            if pg.n % 2 == 0:
                c2ps = generate_C2p(pg.n)
                c2pps = generate_C2pp(pg.n)
                c2s = c2ps + c2pps
            else:
                c2s = generate_C2p(pg.n)
                symels.append(Symel("i", None, inversion_matrix()))
            cns = generate_Cn(pg.n)
            sns = generate_Sn(pg.n * 2, S2n=True)
            sigma_ds = generate_sigma_d(pg.n)
            symels = symels + cns + sns + c2s + sigma_ds
        elif pg.subfamily is None:
            cns = generate_Cn(pg.n)
            if pg.n % 2 == 0:
                c2ps = generate_C2p(pg.n)
                c2pps = generate_C2pp(pg.n)
                c2s = c2ps + c2pps
            else:
                c2s = generate_C2p(pg.n)
            symels = symels + cns + c2s
        else:
            raise Exception(argerr)
    elif pg.family == "S":
        if pg.subfamily is None and (pg.n % 2 == 0):
            n = pg.n >> 1
            if n % 2 != 0:
                symels.append(Symel("i", None, inversion_matrix()))
            cns = generate_Cn(n)
            sns = generate_Sn(pg.n, S2n=True)
            symels = symels + cns + sns
        else:
            raise Exception(argerr)
    else:
        if pg.family == "T":
            if pg.subfamily == "h":
                Ths = generate_Th()
                symels = symels + Ths
            elif pg.subfamily == "d":
                Tds = generate_Td()
                symels = symels + Tds
            else:
                Ts = generate_T()
                symels = symels + Ts
        elif pg.family == "O":
            if pg.subfamily == "h":
                Ohs = generate_Oh()
                symels = symels + Ohs
            else:
                Os = generate_O()
                symels = symels + Os
        elif pg.family == "I":
            if pg.subfamily == "h":
                Ihs = generate_Ih()
                symels = symels + Ihs
            else:
                Is = generate_I()
                symels = symels + Is
        else:
            raise Exception(argerr)
    return symels


def generate_Cn(n):
    symels = []
    axis = np.asarray([0, 0, 1])
    #axis = [0 0 1]'
    cn_r = Cn(axis, n)
    for i in range(1, n):
        a, b = reduce(n, i)
        symels.append(Symel(f"C_{a:d}^{b:d}", axis, matrix_power(cn_r,
                                                                 i)))  # Cns
    return symels


def generate_Sn(n, S2n=False):
    symels = []
    axis = np.asarray([0, 0, 1])
    sigma_h = reflection_matrix(axis)
    cn_r = Cn(axis, n)
    if S2n:  # Generating improper rotations for S2n PG
        for i in range(1, n):
            if i % 2 == 0:
                continue
            else:
                a, b = reduce(n, i)
                if a == 2:
                    continue
                else:
                    symels.append(
                        Symel(f"S_{a}^{b}", axis,
                              np.dot(matrix_power(cn_r, i), sigma_h)))
        return symels
    for i in range(1, n):
        a, b = reduce(n, i)
        if b % 2 == 0:
            b += a
        if a == 2:
            continue
        else:
            symels.append(
                Symel(f"S_{a}^{b}", axis, np.dot(matrix_power(cn_r, i),
                                                 sigma_h)))  # Sns
    return symels


def generate_sigma_v(n):
    if n % 2 == 0:
        nsigma_vs = n >> 1
    else:
        nsigma_vs = n
    symels = []
    x_axis = np.asarray([1, 0, 0])  # Orient C2 and sigma_v along x-axis
    z_axis = np.asarray([0, 0, 1])
    rot_mat = Cn(z_axis, n)
    for i in range(nsigma_vs):
        axis = np.cross(np.dot(matrix_power(rot_mat, i), x_axis), z_axis)
        symels.append(Symel(f"sigma_v({i+1})", axis, reflection_matrix(axis)))
    return symels


def generate_sigma_d(n):
    symels = []
    x_axis = np.asarray([1, 0, 0])  # Orient C2 and sigma_v along x-axis
    z_axis = np.asarray([0, 0, 1])
    rot_mat = Cn(z_axis, 2 * n)
    base_axis = np.dot(
        Cn(z_axis, 4 * n),
        x_axis)  # Rotate x-axis by Cn/2 to produce an axis for sigma_d's
    for i in range(n):
        axis = np.cross(np.dot(matrix_power(rot_mat, i), base_axis), z_axis)
        symels.append(Symel(f"sigma_d({i+1})", axis, reflection_matrix(axis)))
    return symels


def generate_C2p(n):
    if n % 2 == 0:
        nn = n >> 1
    else:
        nn = n
    symels = []
    x_axis = np.asarray([1, 0, 0])  # Orient C2 and sigma_v along x-axis
    rot_mat = Cn([0, 0, 1], n)
    for i in range(nn):
        axis = np.dot(matrix_power(rot_mat, i), x_axis)
        symels.append(Symel(f"C_2'({i+1})", axis, Cn(axis, 2)))
    return symels


def generate_C2pp(n):
    nn = n >> 1
    symels = []
    x_axis = np.asarray([1, 0, 0])
    rot_mat = Cn([0, 0, 1], n)
    base_axis = np.dot(Cn([0, 0, 1], 2 * n), x_axis)
    for i in range(nn):
        axis = np.dot(matrix_power(rot_mat, i), base_axis)
        symels.append(Symel(f"C_2''({i+1})", axis, Cn(axis, 2)))
    return symels


def generate_T():
    """
    Generate symmetry elements for the T point group.
    Assume a tetrahedron contained in a cube, then we can easily generate
    the vectors for the rotation elements.

    :rtype: List[molsym.Symel]
    """
    # Generate C3's
    symels = []
    C3_1v = normalize(np.array([1.0, 1.0, 1.0]))
    C3_2v = normalize(np.array([-1.0, 1.0, -1.0]))
    C3_3v = normalize(np.array([-1.0, -1.0, 1.0]))
    C3_4v = normalize(np.array([1.0, -1.0, -1.0]))
    C3list = [C3_1v, C3_2v, C3_3v, C3_4v]
    namelist = ("alpha", "beta", "gamma", "delta")
    for i in range(4):
        C3 = Cn(C3list[i], 3)
        C3s = matrix_power(C3, 2)
        symels.append(Symel(f"C_3({namelist[i]})", C3list[i], C3))
        symels.append(Symel(f"C_3^2({namelist[i]})", C3list[i], C3s))
    # Generate C2's
    C2_x = np.array([1.0, 0.0, 0.0])
    C2_y = np.array([0.0, 1.0, 0.0])
    C2_z = np.array([0.0, 0.0, 1.0])
    C2list = [C2_x, C2_y, C2_z]
    namelist = ["x", "y", "z"]
    for i in range(3):
        C2 = Cn(C2list[i], 2)
        symels.append(Symel(f"C_2({namelist[i]})", C2list[i], C2))
    return symels


def generate_Td():
    """
    Generate symmetry elements for the Td point group.
    Assume a tetrahedron contained in a cube, then we can easily generate
    the vectors for the rotation elements.

    :rtype: List[molsym.Symel]
    """
    symels = generate_T()
    # σd's
    sigma_d_1v = normalize(np.array([1.0, 1.0, 0.0]))
    sigma_d_2v = normalize(np.array([1.0, -1.0, 0.0]))
    sigma_d_3v = normalize(np.array([1.0, 0.0, 1.0]))
    sigma_d_4v = normalize(np.array([1.0, 0.0, -1.0]))
    sigma_d_5v = normalize(np.array([0.0, 1.0, 1.0]))
    sigma_d_6v = normalize(np.array([0.0, 1.0, -1.0]))
    sigmas = [
        sigma_d_1v, sigma_d_2v, sigma_d_3v, sigma_d_4v, sigma_d_5v, sigma_d_6v
    ]
    namelist = ["xyp", "xym", "xzp", "xzm", "yzp", "yzm"]
    for i in range(6):
        sigma_d = reflection_matrix(sigmas[i])
        symels.append(Symel(f"sigma_d({namelist[i]})", sigmas[i], sigma_d))
    # S4's
    S4_1v = np.array([1.0, 0.0, 0.0])
    S4_2v = np.array([0.0, 1.0, 0.0])
    S4_3v = np.array([0.0, 0.0, 1.0])
    S4vlist = [S4_1v, S4_2v, S4_3v]
    namelist = ["x", "y", "z"]
    for i in range(3):
        S4 = Sn(S4vlist[i], 4)
        S43 = matrix_power(S4, 3)
        symels.append(Symel(f"S_4({namelist[i]})", S4vlist[i], S4))
        symels.append(Symel(f"S_4^3({namelist[i]})", S4vlist[i], S43))
    return symels


def generate_Th():
    """
    Generate symmetry elements for the Th point group.
    Assume a tetrahedron contained in a cube, then we can easily generate
    the vectors for the rotation elements.

    :rtype: List[molsym.Symel]
    """
    symels = generate_T()
    # i
    symels.append(Symel("i", None, inversion_matrix()))
    # S6
    S6_1v = normalize(np.array([1.0, 1.0, 1.0]))
    S6_2v = normalize(np.array([-1.0, 1.0, -1.0]))
    S6_3v = normalize(np.array([-1.0, -1.0, 1.0]))
    S6_4v = normalize(np.array([1.0, -1.0, -1.0]))
    S6list = [S6_1v, S6_2v, S6_3v, S6_4v]
    namelist = ["alpha", "beta", "gamma", "delta"]
    for i in range(4):
        S6 = Sn(S6list[i], 6)
        S65 = matrix_power(S6, 5)
        symels.append(Symel(f"S_6({namelist[i]})", S6list[i], S6))
        symels.append(Symel(f"S_6^5({namelist[i]})", S6list[i], S65))
    # 3sigma_h
    sigma_h_xv = np.array([1.0, 0.0, 0.0])
    sigma_h_yv = np.array([0.0, 1.0, 0.0])
    sigma_h_zv = np.array([0.0, 0.0, 1.0])
    sigma_list = [sigma_h_xv, sigma_h_yv, sigma_h_zv]
    namelist = ["x", "y", "z"]
    for i in range(3):
        sigma_h = reflection_matrix(sigma_list[i])
        symels.append(Symel(f"sigma_h({namelist[i]})", sigma_list[i], sigma_h))
    return symels


def generate_O():
    """
    Generate symmetry elements for the O point group.
    Assume operations on a cube.

    :rtype: List[molsym.Symel]
    """
    symels = []
    # C4
    C4_xv = np.array([1.0, 0.0, 0.0])
    C4_yv = np.array([0.0, 1.0, 0.0])
    C4_zv = np.array([0.0, 0.0, 1.0])
    C4list = [C4_xv, C4_yv, C4_zv]
    namelist = ["x", "y", "z"]
    for i in range(3):
        C4 = Cn(C4list[i], 4)
        C42 = matrix_power(C4, 2)
        C43 = matrix_power(C4, 3)
        symels.append(Symel(f"C_4({namelist[i]})", C4list[i], C4))
        symels.append(Symel(f"C_2({namelist[i]})", C4list[i], C42))
        symels.append(Symel(f"C_4^3({namelist[i]})", C4list[i], C43))
    # C3
    C3_1v = normalize(np.array([1.0, 1.0, 1.0]))
    C3_2v = normalize(np.array([1.0, -1.0, 1.0]))
    C3_3v = normalize(np.array([1.0, 1.0, -1.0]))
    C3_4v = normalize(np.array([1.0, -1.0, -1.0]))
    C3list = [C3_1v, C3_2v, C3_3v, C3_4v]
    namelist = ["alpha", "beta", "gamma", "delta"]
    for i in range(4):
        C3 = Cn(C3list[i], 3)
        C32 = matrix_power(C3, 2)
        symels.append(Symel(f"C_3({namelist[i]})", C3list[i], C3))
        symels.append(Symel(f"C_3^2({namelist[i]})", C3list[i], C32))
    # C2
    C2_1v = normalize(np.array([1.0, 0.0, 1.0]))
    C2_2v = normalize(np.array([1.0, 0.0, -1.0]))
    C2_3v = normalize(np.array([1.0, 1.0, 0.0]))
    C2_4v = normalize(np.array([1.0, -1.0, 0.0]))
    C2_5v = normalize(np.array([0.0, 1.0, 1.0]))
    C2_6v = normalize(np.array([0.0, -1.0, 1.0]))

    C2list = [C2_1v, C2_2v, C2_3v, C2_4v, C2_5v, C2_6v]
    namelist = ["xzp", "xzm", "xyp", "xym", "yzp", "yzm"]
    for i in range(6):
        C2 = Cn(C2list[i], 2)
        symels.append(Symel(f"C_2({namelist[i]})", C2list[i], C2))
    return symels


def generate_Oh():
    """
    Generate symmetry elements for the Oh point group.
    Assume operations on a cube.

    :rtype: List[molsym.Symel]
    """
    symels = generate_O()
    symels.append(Symel("i", None, inversion_matrix()))
    # S4 and σh
    S4_xv = np.array([1.0, 0.0, 0.0])
    S4_yv = np.array([0.0, 1.0, 0.0])
    S4_zv = np.array([0.0, 0.0, 1.0])
    S4list = [S4_xv, S4_yv, S4_zv]
    namelist = ["x", "y", "z"]
    for i in range(3):
        S4 = Sn(S4list[i], 4)
        sigma_h = reflection_matrix(S4list[i])
        S43 = matrix_power(S4, 3)
        symels.append(Symel(f"S_4({namelist[i]})", S4list[i], S4))
        symels.append(Symel(f"sigma_h({namelist[i]})", S4list[i], sigma_h))
        symels.append(Symel(f"S_4^3({namelist[i]})", S4list[i], S43))
    # S6
    S6_1v = normalize(np.array([1.0, 1.0, 1.0]))
    S6_2v = normalize(np.array([1.0, -1.0, 1.0]))
    S6_3v = normalize(np.array([1.0, 1.0, -1.0]))
    S6_4v = normalize(np.array([1.0, -1.0, -1.0]))
    S6list = [S6_1v, S6_2v, S6_3v, S6_4v]
    namelist = ["alpha", "beta", "gamma", "delta"]
    for i in range(4):
        S6 = Sn(S6list[i], 6)
        S65 = matrix_power(S6, 5)
        symels.append(Symel(f"S_6({namelist[i]})", S6list[i], S6))
        symels.append(Symel(f"S_6^5({namelist[i]})", S6list[i], S65))
    # C2
    sigma_d_1v = normalize(np.array([1.0, 0.0, 1.0]))
    sigma_d_2v = normalize(np.array([1.0, 0.0, -1.0]))
    sigma_d_3v = normalize(np.array([1.0, 1.0, 0.0]))
    sigma_d_4v = normalize(np.array([1.0, -1.0, 0.0]))
    sigma_d_5v = normalize(np.array([0.0, 1.0, 1.0]))
    sigma_d_6v = normalize(np.array([0.0, -1.0, 1.0]))

    sigma_dlist = [
        sigma_d_1v, sigma_d_2v, sigma_d_3v, sigma_d_4v, sigma_d_5v, sigma_d_6v
    ]
    namelist = ["xzp", "xzm", "xyp", "xym", "yzp", "yzm"]
    for i in range(6):
        sigma_d = reflection_matrix(sigma_dlist[i])
        symels.append(Symel(f"sigma_d({namelist[i]})", sigma_dlist[i], sigma_d))
    return symels


def generate_I():
    """
    Generate symmetry elements for the I point group.

    :rtype: List[molsym.Symel]
    """
    symels = []
    faces, vertices, edgecenters = generate_I_vectors()
    # C5 (face vectors)
    for i in range(6):
        C5 = Cn(faces[i], 5)
        C52 = matrix_power(C5, 2)
        C53 = matrix_power(C5, 3)
        C54 = matrix_power(C5, 4)
        symels.append(Symel(f"C_5({i})", faces[i], C5))
        symels.append(Symel(f"C_5^2({i})", faces[i], C52))
        symels.append(Symel(f"C_5^3({i})", faces[i], C53))
        symels.append(Symel(f"C_5^4({i})", faces[i], C54))

    # C3 (vertex vectors)
    for i in range(10):
        C3 = Cn(vertices[i], 3)
        C32 = matrix_power(C3, 2)
        symels.append(Symel(f"C_3({i})", vertices[i], C3))
        symels.append(Symel(f"C_3^2({i})", vertices[i], C32))

    # C2 (edge vectors)
    for i in range(15):
        C2 = Cn(edgecenters[i], 2)
        symels.append(Symel(f"C_2({i})", edgecenters[i], C2))

    return symels


def generate_Ih():
    """
    Generate symmetry elements for the Ih point group.

    :rtype: List[molsym.Symel]
    """
    symels = generate_I()
    faces, vertices, edgecenters = generate_I_vectors()
    symels.append(Symel("i", None, inversion_matrix()))
    # S10 (face vectors)
    for i in range(6):
        S10 = Sn(faces[i], 10)
        S103 = matrix_power(S10, 3)
        S107 = matrix_power(S10, 7)
        S109 = matrix_power(S10, 9)
        symels.append(Symel(f"S_10({i})", faces[i], S10))
        symels.append(Symel(f"S_10^3({i})", faces[i], S103))
        symels.append(Symel(f"S_10^7({i})", faces[i], S107))
        symels.append(Symel(f"S_10^9({i})", faces[i], S109))

    # S6 (vertex vectors)
    for i in range(10):
        S6 = Sn(vertices[i], 6)
        S65 = matrix_power(S6, 5)
        symels.append(Symel(f"S_6({i})", vertices[i], S6))
        symels.append(Symel(f"S_6^5({i})", vertices[i], S65))

    # σ (edge vectors)
    for i in range(15):
        sigma_i = reflection_matrix(edgecenters[i])
        symels.append(Symel(f"sigma({i})", edgecenters[i], sigma_i))

    return symels


def generate_I_vectors():
    """
    Vectors defining the faces, vertices, and edges of a regular dodecahedron.

    :rtype: (List[NumPy array of shape (3,)], List[NumPy array of shape (3,)], List[NumPy array of shape (3,)])
    """
    #phi = (1+(np.sqrt(5.0)))/2.0
    #phi_i = 1.0/phi
    #faces_i = np.array([[1.0, phi, 0.0],[1.0, -phi, 0.0],[-1.0, phi, 0.0],[-1.0, -phi, 0.0],
    #         [0.0, 1.0, phi],[0.0, 1.0, -phi],[0.0, -1.0, phi],[0.0, -1.0, -phi],
    #         [phi, 0.0, 1.0],[-phi, 0.0, 1.0],[phi, 0.0, -1.0],[-phi, 0.0, -1.0]])
    #vertices_i = np.array([[1.0, 1.0, 1.0],[1.0, 1.0, -1.0],[1.0, -1.0, 1.0],[-1.0, 1.0, 1.0],
    #            [1.0, -1.0, -1.0],[-1.0, 1.0, -1.0],[-1.0, -1.0, 1.0],[-1.0, -1.0, -1.0],
    #            [0.0, phi, phi_i],[0.0, phi, -phi_i],[0.0, -phi, phi_i],[0.0, -phi, -phi_i],
    #            [phi_i, 0.0, phi],[-phi_i, 0.0, phi],[phi_i, 0.0, -phi],[-phi_i, 0.0, -phi],
    #            [phi, phi_i, 0.0],[phi, -phi_i, 0.0],[-phi, phi_i, 0.0],[-phi, -phi_i, 0.0]])
    ## Reorienting vectors such that one face is on the z-axis with "pentagon" pointing at the POSITIVE?  y-axis
    #theta = -np.arccos(phi/np.sqrt(1+(phi**2)))
    #rmat = rotation_matrix(np.array([1, 0, 0]), theta)
    #lf = np.shape(faces_i)[0]
    #l = np.shape(vertices_i)[0]
    #faces = [np.dot(rmat,faces_i[i,:]) for i in range(lf)]
    #vertices = [np.dot(rmat,vertices_i[i,:]) for i in range(l)]
    #faces = np.dot(rmat, faces_i.T).T
    #vertices = np.dot(rmat, vertices_i.T).T

    #l = np.shape(vertices)[0]
    #edglen = 2*phi_i
    #edgecenters = []
    #for i in range(l):
    #    for j in range(i+1,l):#= i+1:l
    #        if np.isclose(distance(vertices[i], vertices[j]), edglen):
    #            v = normalize(vertices[i]+vertices[j])
    #            same = False
    #            for k in edgecenters:
    #                if np.isclose(abs(np.dot(k,v)), 1.0):
    #                    same = True
    #                    break
    #            if not same:
    #                edgecenters.append(v)
    #for (idx,face) in enumerate(faces):
    #    faces[idx] = normalize(face)
    #face_vectors = []
    #same = False
    #for i in faces:
    #    v = normalize(i)
    #    for j in face_vectors:
    #        same = False
    #        if np.isclose(abs(np.dot(v,j)), 1.0):
    #            same = True
    #            break
    #    if not same:
    #        face_vectors.append(v)
    #
    #vertex_vectors = []
    #same = False
    #for i in vertices:
    #    v = normalize(i)
    #    for j in vertex_vectors:
    #        same = False
    #        if np.isclose(abs(np.dot(v,j)), 1.0):
    #            same = True
    #            break
    #    if not same:
    #        vertex_vectors.append(v)
    face_vectors = [
        np.array([0.0, 0.0, 1.0]),
        np.array([0.0, 2 / np.sqrt(5), 1 / np.sqrt(5)]),
        np.array([
            -np.sqrt((5 + np.sqrt(5)) / 10), (5 - np.sqrt(5)) / 10,
            1 / np.sqrt(5)
        ]),
        np.array([
            np.sqrt((5 + np.sqrt(5)) / 10), (5 - np.sqrt(5)) / 10,
            1 / np.sqrt(5)
        ]),
        np.array([
            -np.sqrt((5 - np.sqrt(5)) / 10), -(5 + np.sqrt(5)) / 10,
            1 / np.sqrt(5)
        ]),
        np.array([
            np.sqrt((5 - np.sqrt(5)) / 10), -(5 + np.sqrt(5)) / 10,
            1 / np.sqrt(5)
        ])
    ]
    vertex_vectors = [
        np.array([
            0.0, -np.sqrt((2 * (5 + np.sqrt(5))) / 15),
            np.sqrt((5 - 2 * np.sqrt(5)) / 15)
        ]),
        np.array([
            0.0, -np.sqrt((2 * (5 - np.sqrt(5))) / 15),
            np.sqrt((5 + 2 * np.sqrt(5)) / 15)
        ]),
        np.array([
            -1 / np.sqrt(3),
            np.sqrt((5 + 2 * np.sqrt(5)) / 15),
            np.sqrt((5 - 2 * np.sqrt(5)) / 15)
        ]),
        np.array([
            -1 / np.sqrt(3), -np.sqrt((5 - 2 * np.sqrt(5)) / 15),
            np.sqrt((5 + 2 * np.sqrt(5)) / 15)
        ]),
        np.array([
            1 / np.sqrt(3),
            np.sqrt((5 + 2 * np.sqrt(5)) / 15),
            np.sqrt((5 - 2 * np.sqrt(5)) / 15)
        ]),
        np.array([
            1 / np.sqrt(3), -np.sqrt((5 - 2 * np.sqrt(5)) / 15),
            np.sqrt((5 + 2 * np.sqrt(5)) / 15)
        ]),
        np.array([
            -np.sqrt((3 * np.sqrt(5) + 5) / (6 * np.sqrt(5))), -np.sqrt(
                (5 - np.sqrt(5)) / 30),
            np.sqrt((5 - 2 * np.sqrt(5)) / 15)
        ]),
        np.array([
            np.sqrt((3 * np.sqrt(5) - 5) / (6 * np.sqrt(5))),
            np.sqrt((5 + np.sqrt(5)) / 30),
            np.sqrt((5 + 2 * np.sqrt(5)) / 15)
        ]),
        np.array([
            -np.sqrt((3 * np.sqrt(5) - 5) / (6 * np.sqrt(5))),
            np.sqrt((5 + np.sqrt(5)) / 30),
            np.sqrt((5 + 2 * np.sqrt(5)) / 15)
        ]),
        np.array([
            np.sqrt((3 * np.sqrt(5) + 5) / (6 * np.sqrt(5))), -np.sqrt(
                (5 - np.sqrt(5)) / 30),
            np.sqrt((5 - 2 * np.sqrt(5)) / 15)
        ]),
    ]
    edgecenters = [
        np.array([
            0.0,
            np.sqrt((5 + np.sqrt(5)) / 10), -np.sqrt((5 - np.sqrt(5)) / 10)
        ]),
        np.array([
            0.0,
            np.sqrt((5 - np.sqrt(5)) / 10),
            np.sqrt((5 + np.sqrt(5)) / 10)
        ]),
        np.array([
            0.5, -np.sqrt(1 + (2 / np.sqrt(5))) / 2, -np.sqrt(
                (5 - np.sqrt(5)) / 10)
        ]),
        np.array([
            0.5,
            np.sqrt(1 - (2 / np.sqrt(5))) / 2,
            np.sqrt((5 + np.sqrt(5)) / 10)
        ]),
        np.array([
            0.5, -np.sqrt(1 - (2 / np.sqrt(5))) / 2, -np.sqrt(
                (5 + np.sqrt(5)) / 10)
        ]),
        np.array([
            0.5,
            np.sqrt(1 + (2 / np.sqrt(5))) / 2,
            np.sqrt((5 - np.sqrt(5)) / 10)
        ]),
        np.array([1.0, 0.0, 0.0]),
        np.array([(np.sqrt(5) - 1) / 4, -np.sqrt((5 + np.sqrt(5)) / 10) / 2,
                  np.sqrt((5 + np.sqrt(5)) / 10)]),
        np.array([(np.sqrt(5) - 1) / 4,
                  np.sqrt((5 + np.sqrt(5)) / 10) / 2, -np.sqrt(
                      (5 + np.sqrt(5)) / 10)]),
        np.array([(np.sqrt(5) - 1) / 4, -np.sqrt((5 + np.sqrt(5)) / 2) / 2,
                  0.0]),
        np.array([(np.sqrt(5) - 1) / 4,
                  np.sqrt((5 + np.sqrt(5)) / 2) / 2, 0.0]),
        np.array([(np.sqrt(5) + 1) / 4, -np.sqrt((5 - np.sqrt(5)) / 10) / 2,
                  np.sqrt((5 - np.sqrt(5)) / 10)]),
        np.array([(np.sqrt(5) + 1) / 4,
                  np.sqrt((5 - np.sqrt(5)) / 10) / 2, -np.sqrt(
                      (5 - np.sqrt(5)) / 10)]),
        np.array([(np.sqrt(5) + 1) / 4, -np.sqrt((5 - np.sqrt(5)) / 2) / 2,
                  0.0]),
        np.array([(np.sqrt(5) + 1) / 4,
                  np.sqrt((5 - np.sqrt(5)) / 2) / 2, 0.0]),
    ]

    return (face_vectors, vertex_vectors, edgecenters)


def point_group_classes(point_group, tol=1e-4):
    """
    Given a list of symmetries that form a point
    group. Identify the classes
    point_group : list of symmetries
    """
    # point_group = (nsym,3,3)
    pg_classes = []

    nsym = len(point_group)
    pg_tree = KDTree(point_group.reshape(nsym, -1))

    for isym in range(nsym):
        sym_conjugates = point_group.transpose(
            0, 2, 1) @ point_group[isym][None, :, :] @ point_group
        dd, ii = pg_tree.query(sym_conjugates.reshape(nsym, -1), k=1)
        if len(dd[dd > tol]) != 0:
            exit("Not a point group")
        s1_tmp = set(ii)
        if s1_tmp not in pg_classes:
            pg_classes.append(s1_tmp)

    return [list(item) for item in set(tuple(row) for row in pg_classes)]

def find_axis_angle(Rmat):
    """
    Given an orthogonal matrix 
    find its axes and angle.
    incase of +-np.eye(3), axes is [0,0,0]
    and nfold = 0
    """
    det = np.linalg.det(Rmat)
    A = Rmat * det
    if np.abs(A - np.eye(3)).sum() < 1e-4:
        return 0, np.array([0, 0, 0])
    ###
    w, v = np.linalg.eig(A)
    one_idx = np.argmin(np.abs(w - 1))
    assert np.abs(w[one_idx] - 1) < 1e-4
    axis = v[:, one_idx]
    axis = axis / np.linalg.norm(axis)
    # find angle
    cos_angle = np.trace(A) * 0.5 - 0.5
    Kmat = np.zeros(9, dtype=complex)
    Kmat[1] = -axis[2]
    Kmat[2] = axis[1]
    Kmat[3] = axis[2]
    Kmat[5] = -axis[0]
    Kmat[6] = -axis[1]
    Kmat[7] = axis[0]
    sin_angle = -0.5 * np.trace(Kmat.reshape(3, 3) @ A).real
    if abs(sin_angle) < 1e-3:
        sin_angle = abs(sin_angle)
    angle = np.arctan2(sin_angle, cos_angle)
    return angle, axis


def fix_axis_angle_gauge(axis, nfold):
    """
    Fix the gauge of the provided axis. and changes sign
    of nfold according to that
    The gauge is fixed such that the real part of [z/y/x]
    is real.
    """
    axis_tmp = np.real(axis)
    for i in [2, 1, 0]:
        if abs(axis_tmp[i]) > 1e-4:
            if axis_tmp[i] < 0:
                return -np.array(axis), -nfold
            else:
                return axis, nfold
    return axis, nfold


def find_symm_axis(sym_mats):
    ## find symmmats and nfold degeneracy
    ### in case not found, nfold is returned as None
    """
    Given a list of symmetry matrices that form a point,
    return their determinents, axis and nfold
    for mirrors : nfold == 2
    """
    group_order = len(sym_mats)
    dets = np.linalg.det(sym_mats)

    axes = []
    nfold = []

    for isym in sym_mats:
        ## check if this is S:
        w = np.linalg.eigvals(isym)
        impro_rot = False
        if np.abs(w - 1).min() > 1e-3:
            impro_rot = True
        t, ax = find_axis_angle(isym)
        if impro_rot and abs(t) > 1e-2:
            if t > 0:
                t -= np.pi
            else:
                t += np.pi
        if (abs(t) > 1e-2):
            t = int(np.rint(2 * np.pi / t))
            ax, t = fix_axis_angle_gauge(ax, t)
        else:
            t = 0
        if abs(t) <= 2:
            t = abs(t)
        axes.append(ax)
        nfold.append(t)

    axes = np.array(axes)
    nfold = np.array(nfold, dtype=int)
    return [dets, nfold, axes]

def get_point_grp(symm_mats):
    """
    Identifies the point group label from symmetry matrices.
    Correctly handles D2d vs D2 ambiguity by checking improper axes.
    """
    dets, nfold, axes = find_symm_axis(symm_mats)
    order = len(symm_mats)
    if order == 1: return "C1"
    
    # Identify Proper Rotations
    proper_indices = np.where(np.logical_and(dets > 0, nfold > 0))[0]
    if len(proper_indices) == 0:
        if np.any(dets < 0):
             return "Cs" if np.abs(nfold).max() > 0 else "Ci"
        return "C1"

    nfold_proper = nfold[proper_indices]
    axes_proper = axes[proper_indices]
    n = np.max(nfold_proper)
    
    # --- Principal Axis Selection (Critical for D2d) ---
    candidate_mask = (nfold_proper == n)
    candidate_axes = axes_proper[candidate_mask]
    paxis = candidate_axes[0]
    
    # If multiple max-order axes exist (e.g., D2, D2h, D2d), favor one collinear with S_2n
    if n == 2 and len(candidate_axes) > 1:
        improper_indices = np.where(dets < 0)[0]
        if len(improper_indices) > 0:
            nfold_imp = np.abs(nfold[improper_indices])
            axes_imp = axes[improper_indices]
            # S4 has nfold=4 in rotation part
            if np.any(nfold_imp > n): 
                target_axes = axes_imp[nfold_imp > n]
                for cand in candidate_axes:
                    if np.any(np.abs(np.abs(target_axes @ cand) - 1.0) < 1e-3):
                        paxis = cand
                        break

    # --- Classification ---
    nfold_abs = np.abs(nfold)
    inversion_present = len(nfold_abs[nfold_abs == 0]) > 1
    
    # Cubic/Icosahedral checks
    c3_axes = axes[np.logical_and(dets > 0, nfold_abs == 3)]
    # Filter unique C3 axes
    c3_unique = []
    for ax in c3_axes:
        if not any(np.abs(np.dot(u, ax)) > 1-1e-3 for u in c3_unique): c3_unique.append(ax)
        
    if len(c3_unique) > 1:
        has_c5 = np.any(np.logical_and(dets > 0, nfold_abs == 5))
        has_c4 = np.any(np.logical_and(dets > 0, nfold_abs == 4))
        if has_c5: return "Ih" if inversion_present else "I"
        if has_c4: return "Oh" if inversion_present else "O"
        if inversion_present: return "Th"
        return "Td" if np.any(np.logical_and(dets < 0, nfold_abs == 4)) else "T"

    # Axial Checks
    perp_c2_axes = []
    c2_candidates = axes[np.logical_and(dets > 0, nfold_abs == 2)]
    for ax in c2_candidates:
        if np.abs(np.dot(ax, paxis)) < 1e-3:
            if not any(np.abs(np.dot(u, ax)) > 1-1e-3 for u in perp_c2_axes):
                perp_c2_axes.append(ax)
    nC_perp = len(perp_c2_axes)
    
    mirror_axes = axes[np.logical_and(dets < 0, nfold_abs == 2)]
    has_sigma_h = any(np.abs(np.abs(np.dot(ax, paxis)) - 1.0) < 1e-3 for ax in mirror_axes)
    nsig_v = sum(1 for ax in mirror_axes if np.abs(np.dot(ax, paxis)) < 1e-3)

    if n == nC_perp: # D groups
        if has_sigma_h: return "D%dh" % n
        if nsig_v > 0: return "D%dd" % n
        return "D%d" % n
    else: # C/S groups
        if has_sigma_h: return "C%dh" % n
        if nsig_v > 0: return "C%dv" % n
        if np.any(np.logical_and(dets < 0, nfold_abs == 2 * n)): return "S%d" % (2 * n)
        return "C%d" % n



def generate_symel_to_class_map(symels, ctab):
    """
    Deprecated. Generate a map of the symels to their corresponding classes.
    """
    pg = PointGroup.from_string(ctab.name)
    if pg.n is not None:
        ns = pg.n >> 1  # pg.n floor divided by 2
    ncls = len(ctab.classes)
    nsymel = len(symels)
    class_map = np.zeros(nsymel, dtype=np.int32)
    class_map[0] = 0  # E is always first
    if pg.family == "C":
        if pg.subfamily == "s" or pg.subfamily == "i":
            class_map[1] = 1
        elif pg.subfamily == "h":
            if pg.n % 2 == 0:
                class_map[3:pg.n + 2] = [i for i in range(1, pg.n)
                                        ]  # [2:pg.n] # C_n
                class_map[2] = pg.n  # i
                class_map[1] = pg.n + ns  # σh
                for i in range(pg.n + 2, 2 * pg.n):  # = pg.n+3:2*pg.n # S_n
                    if i > 3 * ns:
                        class_map[i] = i - ns
                    else:
                        class_map[i] = i + ns - 1
            else:
                for i in range(1, pg.n):  # = 2:pg.n+1 # C_n
                    class_map[i + 1] = i
                class_map[1] = pg.n  # σh
                for i in range(pg.n + 1, 2 * pg.n):  # = pg.n+2:2*pg.n # S_n
                    class_map[i] = i
        elif pg.subfamily == "v":
            # The last class is σv (and then σd if n is even), and the last symels are also these!
            cn_class_map(class_map, pg.n, 0, 0)
            if pg.n % 2 == 0:
                class_map[-pg.n:-ns] = ncls - 2
                class_map[-ns:] = ncls - 1
            else:
                class_map[-pg.n:] = ncls - 1
        else:
            class_map[1:] = [i for i in range(1, nsymel)]  # 2:nsymel
    elif pg.family == "S":
        if pg.n % 4 == 0:
            for i in range(1, pg.n):  # = 2:pg.n
                if i <= ns - 1:
                    class_map[i] = 2 * i
                else:
                    class_map[i] = 2 * (i - ns) + 1
        else:
            class_map[1] = ns  # i
            class_map[2:ns + 1] = [i for i in range(1, ns)]  # 2:ns # C_n
            for i in range(ns + 1, pg.n):  # = ns+2:pg.n # S_n
                if i > ns + (pg.n >> 2):
                    class_map[i] = i - (pg.n >> 2)
                else:
                    class_map[i] = i + (pg.n >> 2)
    elif pg.family == "D":
        if pg.subfamily == "h":
            if pg.n % 2 == 0:
                class_map[1] = ncls - 3  # σh
                class_map[2] = (ncls >> 1)  # i
                cn_class_map(class_map, pg.n, 2, 0)  # Cn
                class_map[pg.n + 2:3 * ns + 2] = ns + 1  # C2'
                class_map[3 * ns + 2:2 * pg.n + 2] = ns + 2  # C2''
                for i in range(2 * pg.n + 2,
                               3 * pg.n):  # = 2*pg.n+3:3*pg.n+1 # Sn
                    if i > 3 * pg.n - ns:
                        class_map[i] = i - 2 * pg.n + 3
                    else:
                        class_map[i] = 3 * pg.n + 4 - i
                # The result of C2'×i changes depending if pg.n ≡ 0 (mod 4)
                # but also D2h doesn't need to be flipped because I treated it special
                if pg.n % 4 == 0 or pg.n == 2:
                    class_map[-pg.n:-ns] = ncls - 2  # σv
                    class_map[-ns:] = ncls - 1  # σd
                else:
                    class_map[-pg.n:-ns] = ncls - 1  # σv
                    class_map[-ns:] = ncls - 2  # σd
            else:
                class_map[1] = (ncls >> 1)
                cn_class_map(class_map, pg.n, 1, 0)
                class_map[pg.n + 1:2 * pg.n + 1] = ns + 1
                cn_class_map(class_map, pg.n, 2 * pg.n, ns + 2)
                class_map[-1 - pg.n + 1:] = ncls - 1
        elif pg.subfamily == "d":
            if pg.n % 2 == 0:
                cn_class_map(class_map, pg.n, 0, 0)  # Cn
                class_map[1:pg.n] = 2 * class_map[
                    1:pg.n]  # 2*class_map[2:pg.n].-1 # Reposition Cn
                cn_class_map(class_map, pg.n + 1, pg.n - 1, 0)  # Sn
                class_map[pg.n:2 * pg.n] = 2 * class_map[pg.n:2 * pg.n] - 1
                # 2*(class_map[pg.n+1:2*pg.n].-1) # Reposition Sn
                class_map[-2 * pg.n:-pg.n] = ncls - 2  # C2'
                class_map[-pg.n:] = ncls - 1  # σd
            else:
                class_map[1] = ncls >> 1  # i
                cn_class_map(class_map, pg.n, 1, 0)  # Cn
                for i in range(pg.n + 1, 2 * pg.n):  # = pg.n+2:2*pg.n # Sn
                    if i > pg.n + ns:
                        class_map[i] = i + 2 - pg.n
                    else:
                        class_map[i] = 2 * pg.n + 2 - i
                class_map[-2 * pg.n:-pg.n] = ns + 1
                class_map[-pg.n:] = ncls - 1  # σd
        else:
            cn_class_map(class_map, pg.n, 0, 0)  # Cn
            if pg.n % 2 == 0:
                class_map[-pg.n:-ns] = ncls - 2  # Cn'
                class_map[-ns:] = ncls - 1  # Cn''
            else:
                class_map[-pg.n:] = ncls - 1  # Cn
    else:
        if pg.family == "T":
            if pg.subfamily == "h":
                class_map = np.array([
                    0, 1, 2, 1, 2, 1, 2, 1, 2, 3, 3, 3, 4, 5, 6, 5, 6, 5, 6, 5,
                    6, 7, 7, 7
                ])
            elif pg.subfamily == "d":
                class_map = np.array([
                    0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 4, 4, 4, 4, 4, 4, 3, 3,
                    3, 3, 3, 3
                ])
            else:
                class_map = np.array([0, 1, 2, 1, 2, 1, 2, 1, 2, 3, 3, 3])
        elif pg.family == "O":
            if pg.subfamily == "h":
                class_map = np.array([
                    0, 3, 4, 3, 3, 4, 3, 3, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2,
                    2, 2, 2, 2, 5, 6, 8, 6, 6, 8, 6, 6, 8, 6, 7, 7, 7, 7, 7, 7,
                    7, 7, 9, 9, 9, 9, 9, 9
                ])
            else:
                class_map = np.array([
                    0, 1, 2, 1, 1, 2, 1, 1, 2, 1, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4,
                    4, 4, 4, 4
                ])
        elif pg.family == "I":
            if pg.subfamily == "h":
                class_map = np.array([
                    0, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2,
                    1, 1, 2, 2, 1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
                    3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
                    5, 6, 7, 7, 6, 6, 7, 7, 6, 6, 7, 7, 6, 6, 7, 7, 6, 6, 7, 7,
                    6, 6, 7, 7, 6, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
                    8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9
                ])
            else:
                class_map = np.array([
                    0, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2,
                    1, 1, 2, 2, 1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
                    3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4
                ])
        else:
            raise Exception(f"An invalid point group has been " +
                            "given or unexpected parsing of the " +
                            "point group string has occured: {pg.str}")
    return class_map


def cn_class_map(class_map, n, idx_offset, cls_offset):
    """
    Deprecated
    """
    for i in range(1, n):  # = 2:n
        if i > (n >> 1):
            class_map[i + idx_offset] = n - i + cls_offset
        else:
            class_map[i + idx_offset] = i + cls_offset
    return class_map






### ============ Unit test ======================
### Generated using chatgpt ####################
### =============================================
if __name__ == '__main__':
    import unittest
    import sys

    ntrials = 50
    class TestPointGroupRobustness(unittest.TestCase):

        def setUp(self):
            # List of groups to test (Crystallographic + Icosahedral)
            self.test_groups = [
                # Triclinic
                "C1", "Ci",
                # Monoclinic
                "C2", "Cs", "C2h",
                # Orthorhombic
                "D2", "C2v", "D2h",
                # Tetragonal
                "C4", "S4", "C4h", "D4", "C4v", "D2d", "D4h",
                # Trigonal
                "C3", "S6", "D3", "C3v", "D3d",
                # Hexagonal
                "C6", "C3h", "C6h", "D6", "C6v", "D3h", "D6h",
                # Cubic
                "T", "Th", "Td", "O", "Oh"
            ]

            self.num_trials = ntrials


        def test_get_pg_info_robustness(self):
            """
            Tests get_pg_info by rotating the molecule randomly in 3D space.
            """
            print("\n")
            print("="*50)
            print("="*20 + "Test 1" + "="*20)
            print("="*50)
            print(f"\nRunning Robustness Test on {len(self.test_groups)} Point Groups")
            print(f"Configurations per group: {self.num_trials}")
            print("=" * 70)

            for label in self.test_groups:
                with self.subTest(group=label):
                    # ----------------------------------------------------------
                    # 1. Generate Ground Truth (Standard Orientation)
                    # ----------------------------------------------------------
                    try:
                        symels = pg_to_symels(label)
                        # Use float64 for precision
                        std_mats = np.array([s.rrep for s in symels], dtype=np.float64)
                    except Exception as e:
                        self.fail(f"[{label}] Setup Error: Could not generate standard symels. {e}")

                    # ----------------------------------------------------------
                    # 2. Run Random Trials
                    # ----------------------------------------------------------
                    for i in range(self.num_trials):
                        R, _ = np.linalg.qr(np.random.randn(3, 3))
                        transformed_mats = np.einsum('ij,njk,kl->nil', R, std_mats, R.T,optimize=True)
                        np.random.shuffle(transformed_mats)
                        mats_clean = []
                        for m in transformed_mats:
                            U, S, Vt = np.linalg.svd(m)
                            m_clean = U @ Vt
                            if np.linalg.det(m_clean) * np.linalg.det(m) < 0:
                                m_clean = -m_clean
                            mats_clean.append(m_clean)
                        transformed_mats = np.array(mats_clean)

                        try:
                            res_label, classes, class_dict, char_tab, irreps = get_pg_info(transformed_mats)
                        except AssertionError as e:
                            # This catches "assert len(pg_sym_mats) == order" and "assert distance < 1e-4"
                            self.fail(f"[{label}] Trial {i+1} FAILED: Library Assertion Error. "
                                  f"The algorithm failed to identify or align the rotated group. "
                                  f"\nMatrix Det: {np.linalg.det(R):.2f}")
                        except Exception as e:
                            self.fail(f"[{label}] Trial {i+1} CRASHED: {e}")

                        self.assertEqual(res_label, label,
                                     f"[{label}] Trial {i+1}: Misclassified as '{res_label}'")

                        if char_tab.ndim == 1:
                            ct_rows, ct_cols = 1, char_tab.shape[0]
                        else:
                            ct_rows, ct_cols = char_tab.shape

                        self.assertEqual(len(irreps), ct_rows,
                                     f"[{label}] CharTable rows {ct_rows} != Irreps {len(irreps)}")
                        self.assertEqual(len(classes), ct_cols,
                                     f"[{label}] CharTable cols {ct_cols} != Classes {len(classes)}")

                        assigned_indices = []
                        for idx_list in class_dict.values():
                            assigned_indices.extend(idx_list)

                        self.assertEqual(len(assigned_indices), len(std_mats),
                                     f"[{label}] Class dictionary missing matrices")
                        self.assertEqual(len(set(assigned_indices)), len(std_mats),
                                     f"[{label}] Class dictionary contains duplicates")

            print("\nSUCCESS: All groups passed Test 1")

    class SymmetryBaseXX:
        """
        Port of the symmetry definitions from Quantum Espresso's symm_base.f90.
        """
        def __init__(self):
            # Constants
            sin3 = 0.866025403784438597
            cos3 = 0.5
            msin3 = -0.866025403784438597
            mcos3 = -0.5

            # Raw data from symm_base.f90
            # This is a flat list of the 32 matrices (3x3).
            # Fortran DATA statements fill arrays in column-major order.
            raw_s0 = [
                # 1-9 (Identity, 180s, 90s)
                1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0,   # 1: Identity
                -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0, 1.0, # 2: 180 z
                -1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, -1.0, # 3: 180 y
                1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0, -1.0, # 4: 180 x
                0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, -1.0,  # 5: 180 [1,1,0]
                0.0, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, -1.0,# 6: 180 [1,-1,0]
                0.0, -1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0,  # 7: 90 z
                0.0, 1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 1.0,  # 8: -90 z

                # 10-17 (More 180s, 90s)
                0.0, 0.0, 1.0, 0.0, -1.0, 0.0, 1.0, 0.0, 0.0,  # 9: 180 [1,0,1]
                0.0, 0.0, -1.0, 0.0, -1.0, 0.0, -1.0, 0.0, 0.0,# 10: 180 [-1,0,1]
                0.0, 0.0, 1.0, 0.0, 1.0, 0.0, -1.0, 0.0, 0.0,  # 11: 90 y
                0.0, 0.0, -1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0,  # 12: -90 y
                0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0,   # 13: 180 [0,1,1]
                0.0, 0.0, -1.0, -1.0, 0.0, 0.0, 0.0, 1.0, 0.0, # 14: 180 [0,1,-1]
                1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, -1.0, 0.0,  # 15: 90 x
                1.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, 1.0, 0.0,  # 16: -90 x

                # 18-25 (Cubic Diagonals / 120 deg)
                0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0,   # 17: 120 [-1,-1,-1]
                0.0, -1.0, 0.0, 0.0, 0.0, 1.0, -1.0, 0.0, 0.0, # 18: 120 [-1,1,1]
                0.0, -1.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, # 19: 120 [1,1,-1]
                0.0, 1.0, 0.0, 0.0, 0.0, -1.0, -1.0, 0.0, 0.0, # 20: 120 [1,-1,1]
                0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0,   # 21: 120 [1,1,1]
                0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 0.0, -1.0, 0.0, # 22: 120 [-1,1,-1]
                0.0, 0.0, -1.0, -1.0, 0.0, 0.0, 0.0, 1.0, 0.0, # 23: 120 [1,-1,-1]
                0.0, 0.0, 1.0, -1.0, 0.0, 0.0, 0.0, -1.0, 0.0, # 24: 120 [-1,-1,1]

                # 26-29 (Hexagonal 60/120 z)
                cos3, msin3, 0.0, sin3, cos3, 0.0, 0.0, 0.0, 1.0,   # 25: 60 z
                cos3, sin3, 0.0, msin3, cos3, 0.0, 0.0, 0.0, 1.0,   # 26: -60 z
                mcos3, msin3, 0.0, sin3, mcos3, 0.0, 0.0, 0.0, 1.0, # 27: 120 z
                mcos3, sin3, 0.0, msin3, mcos3, 0.0, 0.0, 0.0, 1.0, # 28: -120 z

                # 30-32 (Hexagonal 180s)
                cos3, msin3, 0.0, msin3, mcos3, 0.0, 0.0, 0.0, -1.0,# 29: 180 [1,-1,0]
                cos3, sin3, 0.0, sin3, mcos3, 0.0, 0.0, 0.0, -1.0,  # 30: 180 [2,1,0]
                mcos3, msin3, 0.0, msin3, cos3, 0.0, 0.0, 0.0, -1.0,# 31: 180 [0,1,0]
                mcos3, sin3, 0.0, sin3, cos3, 0.0, 0.0, 0.0, -1.0   # 32: 180 [1,1,0]
            ]

            # Reshape: (32, 3, 3).
            # Note: Fortran fills columns first. A reshape(32,3,3) in numpy fills rows.
            # We must reshape to (3,3,32) with Fortran order, then transpose axes.
            s0_fortran = np.array(raw_s0).reshape((3, 3, 32), order='F')

            # Move the 32 axis to the front: (32, 3, 3)
            self.s0 = np.moveaxis(s0_fortran, -1, 0)

            # Create Inversions (Indices 32-63)
            self.s_full = np.concatenate((self.s0, -self.s0), axis=0)

        def get_group_by_generators(self, gen_indices):
            """
            Constructs a full group by closing the set of generators.
            gen_indices: List of indices into s_full (0-63).
            """
            generators = [self.s_full[i] for i in gen_indices]

            # Closure: Keep multiplying until set size is constant
            group = list(generators)

            # Add Identity explicitly if not present (Index 0)
            id_mat = self.s_full[0]
            if not any(np.allclose(g, id_mat, atol=1e-5) for g in group):
                group.append(id_mat)

            while True:
                current_len = len(group)
                new_mats = []
                for m1 in group:
                    for m2 in group:
                        prod = m1 @ m2
                        # Check if prod exists in group
                        is_new = True
                        for existing in group:
                            if np.allclose(prod, existing, atol=1e-5):
                                is_new = False
                                break
                        if is_new:
                            # Also check against new_mats to avoid duplicates in this batch
                            for nm in new_mats:
                                if np.allclose(prod, nm, atol=1e-5):
                                    is_new = False
                                    break
                        if is_new:
                            new_mats.append(prod)

                if not new_mats:
                    break
                group.extend(new_mats)

            return np.array(group)

    class TestAll32Groups(unittest.TestCase):
        @classmethod
        def setUpClass(cls):
            cls.qe = SymmetryBaseXX()

            # =========================================================================
            # Map 32 Point Groups 
            # 0-31: Proper Rotations | 32-63: Inversions (Inv * s0)
            # =========================================================================
            I = 32 # Offset for inversion

            # Indices derived from symm_base.f90 comments
            # 0: E
            # 1: C2z, 2: C2y, 3: C2x
            # 6: C4z (90 z), 7: C4z^-1
            # 24: C6z (60 z) -> s0 index 25 in Fortran (1-based) -> 24 Python
            # 26: C3z (120 z)
            # 16: C3 diagonal (120 [-1,-1,-1])

            cls.generators = {
                # Triclinic
                "C1": [0],
                "Ci": [0 + I],

                # Monoclinic (Standard: z-axis)
                "C2": [1],       # C2z
                "Cs": [1 + I],   # sigma_h (Inv * C2z)
                "C2h": [1, 0+I], # C2z, i

                # Orthorhombic
                "D2": [1, 3],      # C2z, C2x
                "C2v": [1, 3 + I], # C2z, sigma_v(x) (Inv * C2x)
                "D2h": [1, 3, 0+I],# C2z, C2x, i

                # Tetragonal
                "C4": [6],         # C4z
                "S4": [6 + I],     # S4z (Inv * C4z)
                "C4h": [6, 0+I],   # C4z, i
                "D4": [6, 3],      # C4z, C2x
                "C4v": [6, 3 + I], # C4z, sigma_v(x)
                "D2d": [6 + I, 3], # S4z, C2x (Classic D2d generators)
                "D4h": [6, 3, 0+I],

                # Trigonal (Using Hexagonal basis indices 24-31)
                "C3": [26],           # C3z (120 deg)
                "S6": [26, 0 + I],    # C3z, i (S6 = C3 x i)
                "D3": [26, 28],       # C3z, C2' (180 perp)
                "C3v": [26, 28 + I],  # C3z, sigma_v
                "D3d": [26, 28, 0+I], # C3z, C2', i

                # Hexagonal
                "C6": [24],           # C6z
                "C3h": [26, 1 + I],   # C3z, sigma_h (Inv * C2z is sigma_h perp to z)
                "C6h": [24, 0 + I],   # C6z, i
                "D6": [24, 28],       # C6z, C2'
                "C6v": [24, 28 + I],  # C6z, sigma_v
                "D3h": [26, 28, 1+I], # C3z, C2', sigma_h
                "D6h": [24, 28, 0+I],

                # Cubic
                "T": [3, 16],      # C2x, C3_diag
                "Th": [3, 16, 0+I],# T + i
                "O": [6, 16],      # C4z, C3_diag
                "Td": [16, 6 + I], # C3_diag, S4z
                "Oh": [6, 16, 0+I] # O + i
            }

        def test_randomized_32_groups(self):
            """
            Generates all 32 groups and applies random rotations,
            and verifies detection.
            """
            n_trials = ntrials # Number of random rotations per group

            print("\n")
            print("="*50)
            print("="*20 + "Test 2" + "="*20)
            print("="*50)
            print(f"Testing All 32 Point Groups with {n_trials} Random Rotations")
            print(f"{'='*70}")

            for label, gen_idx in self.generators.items():
                with self.subTest(group=label):
                    # 1. Generate Standard Group
                    mats = self.qe.get_group_by_generators(gen_idx)

                    # Verify Order (Sanity Check)
                    # print(f"Testing {label:<4} (Order {len(mats)})")

                    for i in range(n_trials):
                        # 2. Random Rotation
                        # qr decomposition of random matrix gives a uniform random rotation
                        R, _ = np.linalg.qr(np.random.randn(3, 3))

                        # Apply rotation: R * M * R.T
                        mats_rot = np.einsum('ij,njk,kl->nil', R, mats, R.T)

                        # Clean numerical noise: Force matrices to be perfectly orthogonal
                        mats_clean = []
                        for m in mats_rot:
                            U, S, Vt = np.linalg.svd(m)
                            # Reconstruct as a perfect rotation matrix (U * Vt)
                            m_clean = U @ Vt
                            # Fix determinant sign (ensure it doesn't flip improper/proper)
                            # (Ideally SVD preserves it, but checking det is safer)
                            if np.linalg.det(m_clean) * np.linalg.det(m) < 0:
                                m_clean = -m_clean
                            mats_clean.append(m_clean)
                        mats_rot = np.array(mats_clean)

                    # 3. Detect
                        try:
                            detected, _, _, _, _ = get_pg_info(mats_rot)
                        except Exception as e:
                            self.fail(f"[{label}] Crash on trial {i}: {e}")

                        # 4. Assert
                        # Note: S6 is often detected as C3i (equivalent)
                        #       C3h often C6... no, C3h is distinct.
                        #       Cs can be Cs or C1h (synonyms).

                        expected = label
                        # Handle common synonyms in output
                        if label == "S6" and detected == "C3i": detected = "S6"
                        if label == "C3i" and detected == "S6": detected = "C3i"
                        if label == "Cs" and detected == "C1h": detected = "Cs"

                        self.assertEqual(detected, expected,
                                         f"Trial {i}: Expected {expected}, got {detected}. "
                                         f"\n(Group Order: {len(mats)})")

            print("\nSUCCESS: All 32 groups passed Test 2")


    unittest.main(verbosity=2)




from yambopy import *
import sys
import argparse

if __name__ == "__main__":
    #parse options
    parser = argparse.ArgumentParser(description='RT Time step optimization')
    parser.add_argument('-F', '--input_file',type=str,help='<Required> RT input file',required=True)
    parser.add_argument('-D', '--directory', type=str,help='RT conv directory',required=False)

    args = parser.parse_args()

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    yambo_rt = '/Users/fulvio.paleari/software/yambo-devel/bin/yambo_rt' 
    SAVE_path = './database/FixSymm/SAVE'    
    #YamboRTStep_Optimize(input_path=args.input_file,SAVE_path=SAVE_path)
    #YamboRTStep_Optimize(input_path=args.input_file,SAVE_path=SAVE_path,RUN_path=args.directory,TStep_MAX=80,TStep_increase=20,NSimulations=4,ref_time=60,tol_pol=5e-3)
    #YamboRTStep_Optimize(input_path=args.input_file,SAVE_path=SAVE_path,RUN_path=args.directory,TStep_MAX=12,TStep_increase=2,NSimulations=6,ref_time=60,tol_pol=5e-3)
    YamboRTStep_Optimize(input_path=args.input_file,SAVE_path=SAVE_path,RUN_path=args.directory,TStep_MAX=12,TStep_increase=2,NSimulations=6,yambo_rt=yambo_rt)
    #YamboRTStep_Optimize(input_path=args.input_file,SAVE_path=SAVE_path,RUN_path=args.directory,TStep_MAX=10,TStep_increase=2.5,NSimulations=4,yambo_rt=yambo_rt)



#
# Author: Henrique Pereira Coutada Miranda
# Run a GW calculation using Yambo
#
from __future__ import print_function
from yambopy import *
from qepy import *
from schedulerpy import *

yambo =  'yambo'
scheduler = Scheduler.factory

#check if the nscf cycle is present
if os.path.isdir('nscf/bn.save'):
    print('nscf calculation found!')
else:
    print('nscf calculation not found!')
    exit()

#check if the SAVE folder is present
if not os.path.isdir('database/SAVE'):
    print('preparing yambo database')
    p2y_run = scheduler()
    p2y_run.add_command('mkdir -p database')    
    p2y_run.add_command('cd nscf/bn.save; p2y > p2y.log')
    p2y_run.add_command('yambo > yambo.log')
    p2y_run.add_command('mv SAVE ../../database')
    p2y_run.run()

if not os.path.islink('gw/SAVE'):
    s = scheduler()
    s.add_command('mkdir -p gw')
    s.add_command('cd gw; ln -s ../database/SAVE')
    s.run()

#create the yambo input file
y = YamboIn.from_runlevel('%s -p p -g n -V all'%yambo,folder='gw')
    
y['EXXRLvcs'] = [60,'Ry']       # Self-energy. Exchange
y['BndsRnXp'] = [1,40]          # Screening. Number of bands
y['NGsBlkXp'] = [3,'Ry']        # Cutoff Screening
y['GbndRnge'] = [1,30]          # Self-energy. Number of bands

#read values from QPkrange
values, units = y['QPkrange']
kpoint_start, kpoint_end, band_start, band_end = values
#set the values of QPkrange
y['QPkrange'] = [kpoint_start,kpoint_end,2,6]

y.write('gw/yambo_run.in')

print('running yambo')
yambo_run = scheduler()
yambo_run.add_command('cd gw; %s -F yambo_run.in -J yambo'%yambo)
yambo_run.run()



#
# Authors: MN
#
import numpy as np

def write_cube(filename, data, lat_vec, atom_pos, atomic_num, origin=np.zeros(3), header=''):
    """
    Generate a Gaussian cube for a real data (Nx, Ny, Nz)
    filename : Name of the file
    data : real data with dimension (Nx, Ny, Nz)
    lat_vec : lattice vectors ai = a[:,i]
    origin : origin of the cell 
    atom_pos: atomic positions in cart units (bohr)
    atomic_num: atomic numbers of elements
    """
    ## atomic positions must be in [0,1)
    apos_crys = atom_pos@np.linalg.inv(lat_vec.T)
    apos_crys = apos_crys-np.floor(apos_crys)
    apos_crys = (apos_crys+1e-6)%1
    apos_new = apos_crys@lat_vec.T

    with open(filename, "w") as cube:
        cube.write("# Cube file generated using YamboPy\n")
        cube.write('# '+header.strip()+'\n')
        cube.write("%d     %.6f     %.6f      %.6f\n" \
                   %(len(atom_pos), origin[0], origin[1],origin[2] ))
        for i in range(3):
            Ntmp = data.shape[i]
            cube.write("%d     %.6f     %.6f      %.6f\n" \
                   %(Ntmp, lat_vec[0,i]/Ntmp, lat_vec[1,i]/Ntmp,lat_vec[2,i]/Ntmp))

        for i in range(len(atom_pos)):
            cube.write("%d     %d     %.6f     %.6f      %.6f\n" \
                   %(atomic_num[i], 0.0, apos_new[i,0], apos_new[i,1], apos_new[i,2]))

        for ix in range(data.shape[0]):
            for iy in range(data.shape[1]):
                for iz in range(data.shape[2]):
                    cube.write("%.6f      " %(data[ix,iy,iz]))
                    if iz % 6 == 5 : cube.write("\n")
                cube.write("\n")




from qepy import *
from yambopy import *
from materials import *

# Path of executables
pw_path = '/Users/fulvio.paleari/software/q-e/bin'
y_path = '/Users/fulvio.paleari/software/yambo-andreaM/bin'

prefix = 'bn'

# List of coarse grids (CG)
#cg_grids = [[3,3,1],[6,6,1],[9,9,1],[12,12,1]]
cg_grids = [[3,3,1],[6,6,1]]
# List of random fine grids (FG)
#fg_grids = [[9,18,36],[36,72,144],[81],[144]]
fg_grids = [[9,18,36],[36]]

# Laser energy (eV)
E_laser = 5.

# Paths of input data: QE scf save, pseudos
input_data = '/Users/fulvio.paleari/software/whypy/yambo-whypy-devel/tutorial/double-grid/input_data'
scf_save_path = input_data
pseudo_path = '%s/pseudos'%input_data
work_dir = '/Users/fulvio.paleari/software/whypy/yambo-whypy-devel/tutorial/double-grid'

# BN inputs
bn_inp = hBN_1l_test(prefix=prefix,pseudo_path=pseudo_path)
qe_input    = bn_inp.nscfin
yambo_input = bn_inp.ipin

#QE output prefix
nscf_out = "nscf" #"slurm"
#Yambo output folder(s):
y_out_dir = "results"

#Call to main class [convergence on]
YamboDG_Optimize(cg_grids,fg_grids,prefix,qe_input,yambo_input,E_laser=E_laser,STEPS='all',converge_DG=True,\
                scf_save_path=scf_save_path,pseudo_path=pseudo_path,RUN_path=work_dir,\
                nscf_out=nscf_out,y_out_dir=y_out_dir,\
                pw_exec_path=pw_path,yambo_exec_path=y_path,yambo_exec='yambo',\
                save_type='simple',yambo_calc_type='ip')



from __future__ import print_function
from yambopy import *
from schedulerpy import *
import argparse
import os

layer_separation = 12
folder = 'gw+bse'
yambo  = 'yambo'
p2y    = 'p2y'
prefix = 'bn'
scheduler = Scheduler.factory

def create_save():
    #check if the nscf cycle is present
    if os.path.isdir('nscf/%s.save'%prefix):
        print('nscf calculation found!')
    else:
        print('nscf calculation not found!')
        exit()

    #check if the SAVE folder is present
    if not os.path.isdir('database'):
        print('preparing yambo database')
        shell = scheduler()
        shell.add_command('cd nscf/%s.save; %s; %s'%(prefix,p2y,yambo))
        shell.add_command('cd ../../')
        shell.add_command('mkdir -p database')
        shell.add_command('mv nscf/%s.save/SAVE database'%prefix)
        shell.run()

    #create the folder to run the calculation
    if not os.path.isdir(folder):
        shell = scheduler()
        shell.add_command('mkdir -p %s'%folder)
        shell.add_command('cp -r database/SAVE %s/'%folder)
        shell.run()

def run(nthreads=1,cut=False):
    """
    run gw+bse calculation using yambo
    """
    y = YamboIn.from_runlevel('%s -X f -g n -V all'%yambo,folder=folder)

    if cut:
        y['CUTGeo'] = 'box z'
        y['CUTBox'] = [0,0,layer_separation-1]

    QPKrange,_ = y['QPkrange']
    startk,endk,startb,endb = QPKrange
    y['QPkrange'] = [startk,endk,3,6]
    y['FFTGvecs'] = [30,'Ry']
    y['EXXRLvcs'] = [80,'Ry']
    y['NGsBlkXd'] = [3,'Ry']
    y['BndsRnXd'] = [1,30]
    y['GbndRnge'] = [1,30]
    y.arguments.append('em1d')
    y.write('%s/yambo_gw.in'%folder)

    print('running gw')
    shell = scheduler()
    shell.add_command('cd %s; mpirun -np %d %s -F yambo_gw.in -J yambo'%(folder,nthreads,yambo))
    shell.run()

    #
    #create the bse input file
    y = YamboIn.from_runlevel('%s -X s -o b -k sex -y d -V all'%yambo,folder=folder)

    if cut:
        y['CUTGeo'] = 'box z'
        y['CUTBox'] = [0,0,layer_separation-1]

    y['FFTGvecs'] = [30,'Ry']
    y['NGsBlkXs'] = [3,'Ry']
    y['BndsRnXs'] = [1,30]
    y['BSEBands'] = [3,6]
    y['BEnSteps'] = 500
    y['BEnRange'] = [[0,8],'eV']
    y['KfnQPdb'] = 'E < yambo/ndb.QP' #Include previously computed quasiparticle energies
    y.arguments.append('WRbsWF')

    y.write('%s/yambo_bse.in'%folder)

    #run the bse calculation using the dielectric function from gw
    print('running bse')
    shell = scheduler()
    shell.add_command('cd %s; mpirun -np %d %s -F yambo_bse.in -J yambo'%(folder,nthreads,yambo))
    shell.run()

if __name__ == "__main__":

    #parse options
    parser = argparse.ArgumentParser(description='Run GW+BSE calculations on BN.')
    parser.add_argument('-r', '--run',        action="store_true", help='Run BSE calculation')
    parser.add_argument('-c', '--cut',        action="store_true", help='Use coulomb truncation')
    parser.add_argument('-t' ,'--nthreads',                        help='Number of threads', default=1)
   
    args = parser.parse_args()
    nthreads = int(args.nthreads)

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    cut = args.cut

    create_save()
    if args.run:    run(nthreads,cut)
    



from qepy import Path,PwXML

# k-points map
npoints = 50
path_kpoints = Path([ [[0.0, 0.0, 0.0 ],'G'],
                      [[0.0, 0.0, 1.0 ],'H'],
                      [[1./2,0.0,1./2.],'N'],
                      [[0.0, 0.0, 0.0 ],'G'],
                      [[1./2, 1./2, 1./2 ],'P'],
                      [[1./2,0.0,1./2. ],'N']], [npoints,npoints,npoints,npoints,npoints])

# Class PwXML. QE database reading
xml = PwXML(prefix='pw',path='bands')

# Class PwXML. QE database reading
xml.plot_eigen(path_kpoints)



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
from __future__ import print_function
import unittest
import os
import numpy as np
from yambopy.io.outputfile import YamboOut

test_path = os.path.join(os.path.dirname(__file__),'..','..','data','refs','gw')

class TestYamboOut(unittest.TestCase):
    """ This class creates the input files for Si and compares them to reference files
    """
    def test_yamboout(self):

        yo = YamboOut(test_path)
        print(yo)
        assert yo.logs == ['l-yambo_em1d_ppa_HF_and_locXC_gw0_CPU_1',
                           'l-yambo_em1d_ppa_HF_and_locXC_gw0_CPU_2']
        assert yo.run == ['r-yambo_em1d_ppa_HF_and_locXC_gw0']
        assert yo.output == ['o-yambo.qp']
        assert sorted(yo.netcdf) == sorted(['ndb.HF_and_locXC','ndb.QP'])

        #pack the data
        yo.pack()

        #verify ndb.QP data
        keys = sorted(yo.files['ndb.QP'].keys())
        assert keys == ['Band', 'E', 'E-Eo', 'Eo', 'Kpoint', 'Kpoint_index', 'Z', 'qp_table', 'type']

        #verify o-*.qp data
        keys = sorted(yo.files['o-yambo.qp'].keys())
        assert keys == ['Band', 'E-Eo', 'Eo', 'K-point', 'Sc|Eo', 'input', 'type']

        #compare data from ndb.QP and o-yambo.qp
        of = yo.files['o-yambo.qp']['Eo']
        qo = yo.files['ndb.QP']['Eo']
        #the difference is larger than 1e-3
        assert np.all(np.isclose(of,qo,atol=5e-3))

if __name__ == "__main__":
    unittest.main() 



# Copyright (c) 2023-2025, Claudio Attaccalite,
#                          Myrta Gruening, Anna Romani
# All rights reserved.
#
# This file is part of the yambopy project
# Calculate linear response from real-time calculations (yambo_nl)
# Modified to allow generalisation
#
import numpy as np
from yambopy.units import ha2ev,fs2aut
from yambopy.nl.external_efield import Divide_by_the_Field, Gaussian_centre
from tqdm import tqdm
from math import floor, ceil, factorial
import sys
import os
from abc import ABC,abstractmethod
from yambopy.nl.nl_analysis import Xn_from_signal
#
# Derived class for pulse signal
#    
class Xn_from_pulse(Xn_from_signal):
    #
    def set_defaults(self):
        EFIELDS = ["QSSIN"]
        if self.efield["name"] not in EFIELDS:
            raise ValueError(f"Invalid electric field for frequency mixing analysis. Expected one of: {EFIELDS}")
        for i_n in range(len(self.pumps)):
            if self.pumps[i_n]["name"] != 'none':
                raise ValueError("This analysis is for one monochromatic field only.")
        if self.solver == '':
            self.solver = 'full'
        dim_ = 1+self.X_order+(self.X_order/2)**2
        D = (int(dim_),int(dim_ -0.25))
        self.out_dim = D[self.X_order%2]
        self.T0 = Gaussian_centre(self.efield)[0]
        self.sigma = Gaussian_centre(self.efield)[1]
        if self.samp_mod== "":
            self.samp_mod="linear"
        return
    
    def build_map(self):
        M = 1+2*self.X_order + int(self.X_order*(self.X_order-1)/2) 
        L = (int(self.X_order/2*((self.X_order/2+1))),int(((self.X_order+1)/2)**2))
        I = np.zeros((M,2),dtype=int) 
        j = 0
        I[j,:] = [0,0]
        for n in range(2,self.X_order+1,2):
            j += 1
            I[j,:] = [n,floor(n/2)]
        for n in range(1,self.X_order+1):
            for m in range(0,ceil(n/2)):
                j += 1
                I[j,:] = [n,m]
                I[j+L[self.X_order%2],:] = [-n,m]
        return I

    def set_sampling(self,ifrq):
        M = int(1+2*self.X_order + int(self.X_order*(self.X_order-1)/2)) 
        if self.nsamp == -1 or self.nsamp < M:
            self.nsamp = M
            print(f'Sampling points set to {self.nsamp}')
        T_period = 2.0 * np.pi / self.freqs[ifrq]
        T_range, out_of_bounds = self.update_time_range(T_period)
        if (out_of_bounds):
            print(f'User range redifined for frequency {self.freqs[ifrq]* ha2ev:.3e} [eV]')
        print(f"Time range: {T_range[0] / fs2aut:.3f} - {T_range[1] / fs2aut:.3f} [fs]")
        return T_range
        
    def update_time_range(self,T_period): 
        T_range = self.T_urange
        out_of_bounds = False
        if T_range[0] <= 0.0:
            T_range[0] = self.T0 - self.sigma
        T_range[1] = T_range[0] + T_period
        if T_range[1] > self.time[-1]:
            T_range[1] = self.time[-1]
            T_range[0] = T_range[1] - T_period
            out_of_bound = True
        return T_range, out_of_bounds

    def define_matrix(self,T_i,ifrq):
        dim_ = 1+2*self.X_order + int(self.X_order*(self.X_order-1)/2) 
        W = self.freqs[ifrq]
        M = np.zeros((self.nsamp, dim_), dtype=np.cdouble)
        i_map = self.build_map()
        M[:,0] = 1.0
        for ii in range(1,len(i_map)):
            i_n,i_m = i_map[ii]
            n = abs(i_n)
            M[:,ii] = np.exp(-n*(T_i[:]-self.T0)**2/(2*self.sigma**2)) #sign
            if (n%2 == 0 and i_m == int(n/2)):
                M[:,ii] = M[:,ii]
            elif (i_m < ceil(n/2)):
                if (i_n>0):
                    M[:,ii] *= np.exp(-1j * (n-2*i_m)*W * T_i[:]) 
                if (i_n<0):
                    M[:,ii] *= np.exp(1j * (n-2*i_m)*W * T_i[:]) 
        return M
    
    def divide_by_factor(self,f,n,m): # note that this must be generalised and moved to divide by field
        #
        # 1/2 (iEo)^n (-1)^m K (-p*omega; omega, ..., omega, -omega, ..., -omega) => specific for one freq.
        #
        # then K = n!/(m!(n-m)!) * 2**(1-n) if p /= 0, omega \=0 
        #      K = 2**(-n) if p = 0, omega \=0
        #      K = 1 p = 0, omega  =0
        p  = int(n - 2*m)
        factor = 0.5
        if self.freqs[f] > 1e-4:
            factor *=np.power(2,-n,dtype=np.cdouble)
            if p != 0:
                factor *= 2*factorial(n)/factorial(m)/factorial(n-m)
        factor *= (-1)**m*np.power(1.0j*self.efield['amplitude'],n,dtype=np.cdouble)
        return 1.0/factor
        
    def output_analysis(self,out,to_file=True):
        i_map = self.build_map()
        for ii in range(len(i_map)):
            i_n,i_m = i_map[ii]
            if (i_n < 0): continue
            i_p = int(i_n - 2 * i_m)
            T = 10000.0
            for i_f in range(self.n_runs):
                    out[ii, i_f, :] *= self.divide_by_factor(i_f,i_n,i_m) 
                    T_period = 2.0 * np.pi / self.freqs[i_f]
                    Trange, _ = self.update_time_range(T_period)
                    T = min(T,Trange[0])
            out[ii,:,:]*=self.get_Unit_of_Measure(i_n) 
            run_info = self.append_runinfo(T)
            if (to_file):
                output_file = f'o{self.prefix}.YPy-X_order_{i_n}_{i_p}'
                header = "E[eV] " + " ".join([f"X{i_n}({i_p}w)/Im({d}) X{i_n}({i_p}w)/Re({d})" for d in ('x','y','z')])
                values = np.column_stack((self.freqs * ha2ev, out[ii, :, 0].imag, out[ii, :, 0].real,
                                out[ii, :, 1].imag, out[ii, :, 1].real,
                                out[ii, :, 2].imag, out[ii, :, 2].real))
                np.savetxt(output_file, values, header=header, delimiter=' ', footer="Pulse analysis results") 
                outf = open(output_file, "a")
                outf.writelines(run_info)
                outf.close()
            else:
                print(run_info)
                return (self.freqs, out) 

    def reconstruct_signal(self,out,to_file=True):
        i_map = self.build_map()
        Seff = np.zeros((self.n_runs, 3, len(self.time)), dtype=np.cdouble)
        for i_f in tqdm(range(self.n_runs)):
            for i_d in range(3):
                for ii in range(len(i_map)):
                    i_n,i_m = i_map[ii]
                    if (i_n < 0): continue
                    i_p = int(i_n - 2 * i_m)
                    freq_term = (np.exp(-1j * i_p * self.freqs[i_f] * self.time))*np.exp(-i_n*(self.time-self.T_0)**2/(2*self.sigma**2))
                    Seff[i_f, i_d, :] += out[i_order, i_f, i_d] * freq_term
                    Seff[i_f, i_d, :] += np.conj(out[i_order, i_f, i_d]) * np.conj(freq_term)
        for i_f in tqdm(range(self.n_runs)):
            values = np.column_stack((self.time / fs2aut, Seff[i_f, 0, :].real, Seff[i_f, 1, :].real, Seff[i_f, 2, :].real))
            output_file = f'o{self.prefix}.YamboPy-pol_reconstructed_F{i_f + 1}'
            header="[fs] Px Py Pz"
            if (to_file):
                np.savetxt(output_file, values, header=header, delimiter=' ', footer="Reconstructed signal")
        if (not to_file):
            return values




#
# Authors: MN
#
import atexit

# Global citation registry
_CITATIONS_USED = set()


def citation(ref: str):
    """
    Decorator that records a citation reference when the function is executed.

    This decorator may be applied multiple times to the same function,
    enabling stacked usage such as::

        @citation("A")
        @citation("B")
        def func():
            ...

    Each reference is recorded when the wrapped function is run. Duplicates
    across multiple calls are ignored.

    Parameters
    ----------
    ref : str
        A reference string identifying the source to cite, such as
        a publication title, DOI, or URL.

    Returns
    -------
    callable
        A decorated function that appends the citation to the global registry
        whenever invoked.
    """
    def wrapper(func):
        def inner(*args, **kwargs):
            _CITATIONS_USED.add(ref)
            return func(*args, **kwargs)

        # Preserve function identity for debugger, help(), and Sphinx autodoc
        inner.__name__ = func.__name__
        inner.__doc__ = func.__doc__
        inner.__qualname__ = func.__qualname__
        return inner
    return wrapper


@atexit.register
def print_citations():
    """
    Print a numbered citation list on interpreter shutdown.

    If any decorated functions were executed during the program run,
    this function prints a formatted list of unique citation references
    in sorted order.

    Notes
    -----
    This function is automatically registered via :mod:`atexit`.
    Users should not call this function directly.

    Examples
    --------
    Typical output:: Run this script.
    """
    if not _CITATIONS_USED:
        return

    print("\n=========================================================")
    print("Please cite the following references in case you wish")
    print("to acknowledge the work done by the Authors.")
    # Always print yambopy citation in the start.
    print("=========================================================")
    print("1) F. Paleari, et al. Yambopy. Zenodo (2025). doi: 10.5281/zenodo.15012963")
    for i, ref in enumerate(sorted(_CITATIONS_USED), start=1):
        print(f"{i+1}) {ref}")
    print("=========================================================")


#### test
if __name__ == "__main__":
    @citation("Ref A: Tenet by C. Nolan")
    def Tenet():
        return 1 + 1

    @citation("Ref B: Intersteller by C. Nolan")
    @citation("Ref C: Oppenheimer by C. Nolan")
    def Intersteller():
        return 1 + 1
    # call functcion
    for i in range (3) :
        Intersteller()
        Tenet()



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: FP, MN
#
# This file is part of the yambopy project
#
import numpy as np
from netCDF4 import Dataset
from yambopy.tools.string import marquee
from yambopy.units import ha2ev
from yambopy.kpoints import build_ktree, find_kpt

def get_string_nc(nc_char):
    """
    Convert nc data type to plain string
    """
    if isinstance(nc_char, np.ndarray):
        if nc_char.dtype.kind == 'S':  # Byte strings (C chars)
            nc_char = nc_char.tobytes().decode('utf-8').strip()
        else:
            nc_char = str(nc_char)  # Fallback for non-string arrays
    elif isinstance(nc_char, bytes):
        nc_char = nc_char.decode('utf-8').strip()
    else:
        nc_char = str(nc_char).strip()
    return nc_char.strip().replace('\0', '')

class LetzElphElectronPhononDB():
    """
    Python class to read the electron-phonon matrix elements from LetzElPhC.

    About LetzElPhC: https://github.com/yambo-code/LetzElPhC
    
    By default it reads the full database g(q,k,m,s,b1,b2) including phonon energies.
    
    - Input: path of ndb.elph
    - Input: read_all (default True), read ph. eigenvectors and el-ph matrix elements
    - Input: div_by_energies (default True), divide el-ph mat. el. by sqrt(2* ph. energies)

    - Usage and main variables: 
    
      :: lph = LetzElphElectronPhononDB(path_ndb_elph)
    
      :: lph.kpoints         #kpoints in cryst. coords. (BZ)
      :: lph.qpoints         #qpoints in crist. coords. (BZ)
      :: lph.ph_energies     #Phonon energies (eV)
      :: lph.ph_eigenvectors #Phonon modes
      :: lph.gkkp            #El-ph matrix elements (by default normalised with ph. energies) [!!!! RYDBERG UNITS !!!!]:
      :: lph.gkkp_sq         #Couplings (square)

    Formats:
    - modes[iq][il][iat][ix]
    - gkkp[iq][ik][il][is][ib1][ib2]
    """

    def __init__(self,filename,read_all=True,div_by_energies=True,verbose=False):

        # Open database
        try: database = Dataset(filename)
        except: raise FileNotFoundError("error opening %s in LetzElphElectronPhononDB"%filename)

        self.filename = filename
        # Read DB dimensions
        self.nb1 = database.dimensions['initial_band'].size
        self.nb2 = database.dimensions['final_band_PH_abs'].size
        self.nm = database.dimensions['nmodes'].size
        self.nat = database.dimensions['atom'].size
        self.nk = database.dimensions['nk'].size
        self.nq = database.dimensions['nq'].size
        self.ns = database.dimensions['nspin'].size
        self.nsym = database.dimensions['nsym_ph'].size
        self.div_by_energies = div_by_energies # if true, the elph store are normalized with 1/(2*w_ph)
        #
        #
        conv   = database['convention'][...].data
        self.convention = get_string_nc(conv)
        kernel = database['kernel'][...].data
        self.kernel     = get_string_nc(kernel)

        #
        # Read DB
        self.kpoints = database.variables['kpoints'][:]
        self.qpoints = database.variables['qpoints'][:]
        self.bands   = database.variables['bands'][:]
        self.kmap    = database.variables['kmap'][:]
        self.ktree   = build_ktree(self.kpoints)
        self.qtree   = build_ktree(self.qpoints)

        # Energy units are in rydberg 
        self.ph_energies = database.variables['FREQ'][...].data*(ha2ev/2.)

        self.check_energies()

        if read_all: 
 
            self.read_eigenmodes(database)
            self.read_elph(database,scale_g_with_ph_energies=div_by_energies)

        for var in database.variables.values():
            if var.name=='elph_mat': self.ncfloat_type=var.dtype 

        database.close()
        
        self.verbose = verbose

        if self.verbose:
            if conv == 'standard':
                print("Convention used in Letzelphc : k -> k+q (standard)")
            else:
                print("Convention used in Letzelphc : k-q -> k (yambo)")

    def check_energies(self):
        """
        Inform the user about unexpected negative frequencies and set them to positive
        """
        indices = np.where(self.ph_energies < 0.)
        warn = False
        for Q in indices[0]:
            for M in indices[1]:
                if Q==0 and M in [0,1,2]:
                    self.ph_energies[Q,M]=0.
                else:
                    warn = True
                    self.ph_energies[Q,M]=np.abs(self.ph_energies[Q,M])
        if warn: print("[WARNING] Found unexpected negative phonon energies, be careful!")

    def read_eigenmodes(self,database):
        """
        Read phonon eigenmodes
        """

        self.ph_eigenvectors = np.zeros([self.nq,self.nm,self.nat,3],dtype=np.complex64)
        #eivs_tmp[qpt][mode][atom][coord][cmplx]
        eivs_tmp = database.variables['POLARIZATION_VECTORS'][:]
        self.ph_eigenvectors = eivs_tmp[:,:,:,:,0] + 1j*eivs_tmp[:,:,:,:,1]

    def read_elph(self,database,scale_g_with_ph_energies=True):
        """
        Read electron-phonon matrix elements
        
        - If scale_g_with_ph_energies they are divided by sqrt(2*ph_E)
        """    
        gkkp_full = np.zeros([self.nq,self.nk,self.nm,self.ns,self.nb1,self.nb2],dtype=np.complex64)
        gkkp_tmp  = database.variables['elph_mat'][:]
        gkkp_full = gkkp_tmp[:,:,:,:,:,:,0]+1j*gkkp_tmp[:,:,:,:,:,:,1]
       
        # Check integrity of elph values
        if np.isnan(gkkp_full).any(): print('[WARNING] NaN values detected in elph database.')
        
        # Scaling with phonon energies
        if scale_g_with_ph_energies: gkkp_full = self.scale_g(gkkp_full) 

        self.gkkp = gkkp_full
        self.gkkp_sq = np.abs(gkkp_full)**2. 

    def scale_g(self,dvscf):
        """
        Normalise matrix elements by the phonon energy as: 
       
        g_qnu = dvscf_qnu/sqrt(2*w_qnu)
        """
        
        g = np.zeros([self.nq,self.nk,self.nm,self.ns,self.nb1,self.nb2],dtype=np.complex64)
        for iq in range(self.nq):
            for inu in range(self.nm): 
                if iq==0 and inu in [0,1,2]: 
                    g[iq,:,inu,:,:,:] = 0. # Remove acoustic branches
                else:
                    ph_E = self.ph_energies[iq,inu]/(ha2ev/2.) # Put back the energies in Rydberg units
                    g[iq,:,inu,:,:,:] = dvscf[iq,:,inu,:,:,:]/np.sqrt(2.*ph_E)
        return g


    def read_iq(self,iq, bands_range=[], database=None, convention='yambo'):
        """
        Reads the electron-phonon matrix elements and phonon eigenvectors for a specific q-point index.

        If the data is already loaded in memory, it returns the corresponding array slice. Otherwise,
        it reads from the database without storing the data in memory.
        
        This function reads data for a single q-point instead of the entire dataset, which is useful
        for handling large databases efficiently.

        Parameters
        ----------
        iq : int
            Index of the q-point.
        bands_range : list, optional
            Specifies the range of bands to read. The start index follows Python indexing (starting from 0),
            and the end index is excluded. If not provided, it defaults to the minimum and maximum bands available.
            Please note that the bands range should be inbetween the ones used in letelphc inputfile.
        database : Dataset, optional
            If provided, the function will use this open dataset instead of opening the file again.
        convention : str, optional
            Defines the convention used for electron-phonon matrix elements.
            - 'yambo': Outputs \<k|dV|k-q>.
            - Any other value: Outputs \<k+q|dV|k>.

        Returns
        -------
        tuple
            A tuple containing:
            - ph_eigenvectors : ndarray
                The phonon eigenvectors.
            - ph_elph_me : ndarray
                The electron-phonon matrix elements with the specified convention.
                ( nk, nm, nspin, initial bnd, final bnd)
        """
        #
        if len(bands_range) == 0:
            bands_range = [min(self.bands)-1,max(self.bands)]
        min_bnd = min(bands_range)
        max_bnd = max(bands_range)
        nbnds = max_bnd - min_bnd
        assert (min_bnd >= min(self.bands)-1)
        assert (max_bnd <= max(self.bands))
        start_bnd_idx = 1+min_bnd - min(self.bands)
        end_bnd = start_bnd_idx + nbnds

        # self.ph_eigenvectors , self.gkkp
        if hasattr(self, 'ph_eigenvectors'):
            ph_eigs = self.ph_eigenvectors[iq]
            eph_mat = self.gkkp[iq, :, :, :, start_bnd_idx:end_bnd, start_bnd_idx:end_bnd ]
        else :
            ## else we load from the file
            close_file = False
            if not database :
                close_file = True
                database = Dataset(self.filename,'r')
            eph_mat = database['elph_mat'][iq, :, :, :, start_bnd_idx:end_bnd, start_bnd_idx:end_bnd, :].data
            # ( nk, nm, nspin, initial bnd, final bnd)
            ph_eigs = database['POLARIZATION_VECTORS'][iq,...].data
            eph_mat = eph_mat[...,0] + 1j*eph_mat[...,1]
            ph_eigs = ph_eigs[...,0] + 1j*ph_eigs[...,1]
            ## normalize with ph_freq
            if self.div_by_energies:
                ph_freq_iq = np.sqrt(2.0*np.abs(self.ph_energies[iq])/(ha2ev/2.))
                if iq >0:
                    ph_freq_iq = 1.0/ph_freq_iq
                    eph_mat *= ph_freq_iq[None,:,None,None,None]
                else:
                    eph_mat[:,:3] = 0.0
                    ph_freq_iq = 1.0/ph_freq_iq[3:]
                    eph_mat[:,3:] *= ph_freq_iq[None,:,None,None,None]

            if close_file :database.close()
        ## output elph matrix elements unit (Ry if div_by_energies else Ry^1.5)
        # ( nk, nm, nspin, initial bnd, final bnd)
        return [ph_eigs, self.change_convention(self.qpoints[iq],eph_mat, convention)]

    def change_convention(self, qpt, elph_iq, convention='yambo'):
        """
        Adjusts the convention of the electron-phonon matrix elements.
    
        Parameters
        ----------
        qpt : ndarray
            The q-point in crystal coordinates.
        elph_iq : ndarray
            The electron-phonon matrix elements.
        convention : str, optional
            Defines the output format:
            - 'yambo': Outputs \<k|dV|k-q>.
            - Any other value: Outputs \<k+q|dV|k>.

        Returns
        -------
        ndarray
            The electron-phonon matrix elements in the desired convention. The returned array is a view, not a copy.
        """
        if convention.strip() != 'yambo': convention = 'standard'
        if self.convention == convention: return elph_iq
        if convention == 'standard': factor = 1.0
        else: factor = -1.0
        idx_q = find_kpt(self.ktree, factor*qpt[None, :] + self.kpoints)
        return elph_iq[idx_q, ...]

    def descreen(self,Z,Zval,atomic_masses):
        """
        Obtain approximate "true" bare matrix elements, removing
        the core electron screening. 

        If using pseudopotentials,
        "bare" nuclei are actually still partially screened by
        core frozen core electrons.

        - Read the docstring of descreen_el_ph for information.

        NOTE: Only works with kernel='bare'.
        """
        if self.kernel.strip()!='bare':
            raise ValueError("[ERROR] descreening can only be applied \
                              to mat. el. computed with kernel='bare'")

        ry2ev = ha2ev/2.
        self.gkkp_bare = descreen_el_ph(self.gkkp,self.ph_energies/ry2ev,self.ph_eigenvectors,Z,Zval,masses=atomic_masses)

    def __str__(self):

        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
            
        app('nqpoints: %d'%self.nq)
        app('nkpoints: %d'%self.nk)
        app('nmodes: %d'%self.nm)
        app('natoms: %d'%self.nat)
        app('nbands: %d %d'%(self.nb1,self.nb2))
        app('convention: %s'%self.convention)
        app('kernel: %s'%self.kernel)

        if self.verbose:

            if hasattr(self, 'ph_eigenvectors'):                 
                app('-----------------------------------')
                for iq in range(self.nq):
                    app('nqpoint %d'%iq)
                    for n,mode in enumerate(self.ph_eigenvectors[iq]):
                        app('mode %d freq: %lf meV'%(n,self.ph_energies[iq,n]*1000.))
                        for a in range(self.nat):
                            app(("%12.8lf "*3)%tuple(mode[a].real))
                    app('-----------------------------------')
            else:
                app('-----------------------------------')
                app('Eigenvectors and matrix elements not loaded')
                app('-----------------------------------')

        return "\n".join(lines)



#
# Authors: MN
import numpy as np
from yambopy.kpoints import build_ktree, find_kpt
from yambopy.tools.function_profiler import func_profile
from yambopy.tools.citations import citation


@func_profile
@citation("M. Nalabothula et al. arXiv:2511.21540 (2025)")
def rotate_exc_wf(Ak, symm_mat_red, kpoints, exe_qpt, dmats, time_rev, ktree=None):
    """
    Rotate the exciton wavefunction Ak using symmetry operations.

    This function applies a symmetry operation to the exciton wavefunction Ak,
    which is represented in the basis of electronic states. The rotation is
    performed using the symmetry matrix in reduced coordinates and the
    corresponding representation matrices.
    Implements Eq. (49) of M. Nalabothula et al. arXiv:2511.21540 (2025)

    Parameters
    ----------
    Ak : array_like
        Exciton wavefunction coefficients with shape (n_exe_states, 1 or 2, nspin, nk, nc, nv).
        1 for TDA and 2 for coupling
    symm_mat_red : array_like
        Symmetry matrix in reduced coordinates with shape (3, 3).
    kpoints : array_like
        K-points in the full Brillouin zone (crystal coordinates) with shape (nk, 3).
    exe_qpt : array_like
        Momentum of the given exciton (q-point) in crystal coordinates with shape (3,).
    dmats : array_like
        Representation matrices for the symmetry operation with shape (nk, nspin, Rk_band, k_band).
    time_rev : bool
        If True, apply time-reversal symmetry to the wavefunction.
    ktree : object, optional
        Pre-built k-point tree for efficient k-point searching. If not provided, it will be built.

    Returns
    -------
    rot_Ak : ndarray
        Rotated exciton wavefunction coefficients with the same shape as Ak.
    """
    # Initialize the rotated Ak array
    rot_Ak = np.zeros(Ak.shape, dtype=Ak.dtype)
    # Check TDA
    tda = True
    if Ak.shape[1] == 2 : tda = False

    ns, nk, nc, nv = Ak.shape[2:]
    # Build a k-point tree if not provided
    if ktree is None: ktree = build_ktree(kpoints)

    # Compute the indices of Rk and Rk - q
    Rkpts = kpoints @ symm_mat_red.T  # Rotated k-points
    k_minus_q = kpoints - exe_qpt[None, :]  # k - q
    idx_Rk = find_kpt(ktree, Rkpts)  # Indices of rotated k-points
    idx_k_minus_q = find_kpt(ktree, k_minus_q)  # Indices of k - q

    # Extract the conduction and valence parts of the representation matrices
    Dcc = dmats[:, :, nv:, nv:].transpose(1,0,2,3)  # Conduction band part
    Dvv = dmats[idx_k_minus_q, :, :nv, :nv].transpose(1,0,2,3).conj()  # Valence band part (conjugated)

    # Apply time-reversal symmetry if required
    Ak_tmp = Ak
    if time_rev: Ak_tmp = Ak.conj()

    # Rotate the Ak wavefunction using the representation matrices
    ## rotate the resonant part
    rot_Ak[:, :1, :, idx_Rk, ...] = ((Dcc[None, ...] @ Ak_tmp[:,0,...])
                                     @ (Dvv.transpose(0, 1, 3, 2)[None, ...])
                                     ).reshape(rot_Ak[:,:1].shape)
    if not tda :
        # Rotate the anti-resonant part
        Dvv = dmats[:, :, :nv, :nv].transpose(1,0,2,3)
        Dcc = dmats[idx_k_minus_q, :, nv:, nv:].transpose(1,0,2,3).conj()
        rot_Ak[:, 1:, :, idx_Rk, ...] = ((Dcc[None, ...] @ Ak_tmp[:,1,...])
                                         @ (Dvv.transpose(0, 1, 3, 2)[None, ...])
                                         ).reshape(rot_Ak[:, 1:].shape)
    return rot_Ak




"""
Tutorial for YamboExcitonDB.

Reading excitonic quantities from ndb.BS_diago_Q1

EDIT the path below to point to the yambo SAVE folder.
"""
save_path='BSE_saves/YAMBO_saves'
bse_path ='BSE_saves/BSE_databases'
from yambopy import YamboLatticeDB,YamboExcitonDB
import numpy as np
import matplotlib.pyplot as plt

if __name__ == "__main__":

    #Customly chosen Q-point
    iQ=0 # 0-> Gamma point, i.e., optical absorption limit

    #                    #
    # Start Yambopy part #
    #                    #

    # Create "lattice" object by reading the ns.db1 database inside the yambo SAVE
    ylat = YamboLatticeDB.from_db_file(filename=save_path+'/SAVE/ns.db1')

    # Read exciton data at Q=iQ
    yexc = YamboExcitonDB.from_db_file(ylat,filename=bse_path+'/ndb.BS_diago_Q1')
    print(yexc)

    # Eigenvalues (exciton energies) and intensities (residuals squared)
    print('\nEquivalent of ypp -e s -b 1: ')
    energies    = yexc.eigenvalues.real
    intensities = yexc.get_intensities().real 
    for i_exc in range(10): print(i_exc+1,' %2.4f'%energies[i_exc],' %2.4f'%intensities[i_exc])
    print('...\n ')

    # Eigenvectors (exciton wave functions)
    print('Eigenvector shape (number of excitons, number of transitions): ')
    print(yexc.eigenvectors.shape,'\n ') 

    # Table (from transition basis to single particle basis)
    print('Transition index t = (kvc) -> Single-particle indices k, v, c ')
    for it, t in enumerate(yexc.table[:10]): print(it, ' -> ', t[0],t[1],t[2])
    print('...')

    #                    #
    # Start Yambopy part #
    #                    #



import numpy as np
import matplotlib.pyplot as plt
from qepy Path,PwXML,Unfolding

# Matplotlib options
fig = plt.figure(figsize=(7,5))
ax = fig.add_axes( [ 0.1, 0.1, 0.4, 0.8 ])

# k-points map
npoints = 30

path_kpoints_pc = Path([ [[0.0,  0.5, 0.0],'M'],
                         [[0.0,  0.0, 0.0],'G'],
                         [[1.0/3.0,  1.0/3.0, 0.0],'K'],
                         [[0.0,  0.5, 0.0],'G']], [int(npoints),int(npoints),int(npoints)])

path_kpoints_sc = Path([ [[0.0,  1.0, 0.0],'M'],
                         [[0.0,  0.0, 0.0],'G'],
                         [[2.0/3.0,  2.0/3.0, 0.0],'K'],
                         [[0.0,  1.0, 0.0],'G']], [int(npoints),int(npoints),int(npoints)])

# Prefix and path to read data
prefix_pc = 'BN_pc'
prefix_sc = 'BN_sc_2x2x2'

path_pc = 'Unfolding/BN_pc'
path_sc = 'Unfolding/BN_sc_2x2x2'

# Reading QE database
pc = PwXML(prefix=prefix_pc,path=path_pc)
sc = PwXML(prefix=prefix_sc,path=path_sc)

# Computing projections of sc values into pc
fold = Unfolding(prefix_pc=prefix_pc,path_pc=path_pc,prefix_sc=prefix_sc,path_sc=path_sc,spin="noncol", band_min = 0, sc_rotated=False,compute_projections=True)

# Plotting unfolded sc band structure
fold.plot_eigen_ax(ax,path=path_kpoints_sc,ylim = (-4.0,10.0))

plt.show()

